(***********************************************************************)
(*                                                                     *)
(*                      Naval War for JoCaml                           *)
(*                                                                     *)
(*          Sylvain Conchon, projet Para, INRIA Rocquencourt           *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* Programme de bataille Navale *)

open Graphics;;
open Array;;


type pos = {x: int ; y:int};;
type coup = Loupe | Touche | Coule | Vide;;
type layer = None | Some of image;; 

let nbpiece = 4 ;; 
let n = 10 and m=10;; (* tableau de NxM *)

(*************************************************************)
(** CONVERSION d'un fichier .ppm (RAW) en color array array **)
(*                                                           *)
(*---------------------------------------------------------- *)
(*                                                           *)
(* getint <in_channel> -> int                                *)
(*        sert de parser du fichier ascii                    *)
(* convert <string> -> <color array array>                   *)
(* convert2 <string * color> -> <color array array>          *)
(* get_tcolor < int * int * int * int * color array array>   *)
(*                   -> <color array array>                  *)
(*************************************************************)      
let rec getint f = 
  try
    let c = input_char f in
    match c with
      'P' | '#' -> input_line f;
        getint f
    | ' ' | '\n'-> getint f
    | _ -> let a = String.make 3 ' ' in 
      a.[0] <- c;
      let c2 = input_char f in
      match c2 with
        ' ' | '\n' -> a.[2] <- a.[0];
          a.[0] <- '0';
          a.[1] <- '0';
          int_of_string a
      | _ -> a.[1] <- c2;
          let c3 = input_char f in
          match c3 with
            ' ' | '\n' ->
              a.[2] <- a.[1];
              a.[1] <- a.[0];
              a.[0] <- '0';
              int_of_string a
          | _ -> a.[2] <- c3; 
              int_of_string a

  with
  | End_of_file -> close_in f;
      
      Pervasives.exit 0;;

(* conversion sans transparence *)
let convert fichier =
  let f=open_in fichier in
  let d  = getint f 
  and  e = getint f 
  and  g = getint f in
  let image = create_matrix e d (rgb 0 0 0) in 
  let c0 = input_byte f in
  for j = 0 to (e-1) do
    for i = 0 to (d-1) do
      let c1 = input_byte f
      and c2 = input_byte f
      and c3 = input_byte f in
      image.(j).(i) <- rgb c1 c2 c3;
    done
  done;
  close_in f;
  image;;



(* conversion avec transparence *)
let convert2 fichier =
  let f=open_in fichier in
  let d  = getint f 
  and  e = getint f 
  and  g = getint f in
  let image = create_matrix e d (rgb 0 0 0) in 
  let c0 = input_byte f in
  for j = 0 to (e-1) do
    for i = 0 to (d-1) do
      let c1 = input_byte f
      and c2 = input_byte f
      and c3 = input_byte f in
      let col = rgb c1 c2 c3 in
      if col = (rgb 255 255 255) then image.(j).(i) <- transp
      else
        image.(j).(i) <- col;
    done
  done;
  close_in f;
  image
    ;;


let get_tcolor y x dy dx source=
  let arendre = create_matrix dx dy white
  in
  begin
    for ix = 0 to (dx-1) do
      for jy = 0 to (dy-1) do
        arendre.(ix).(jy) <- source.(x+ix).(y+jy)
      done
    done;
    arendre
  end;;


let changecol source col1 col2=
  let x = Array.length source 
  and y = Array.length source.(0)
  in
  for ix = 0 to (x-1) do
    for iy = 0 to (y-1) do
      if source.(ix).(iy) = col1 then source.(ix).(iy) <- col2
    done
  done;;


(* Initialise la variable bateau des objets player *)
let init_bateau i = 
  convert2 ("sprites/bateau"^(string_of_int i)^".ppm");;
(* Initialise la variable piece_bateau des objets player *)
let init_piece_bateau i =
  convert2 ("sprites/bat"^(string_of_int i)^".ppm");;


let fond_ecran = convert "sprites/fond.ppm";;
let boom = convert "sprites/boom.ppm";;
let eau = convert "sprites/eau.ppm";;
let coule = convert "sprites/coule.ppm";;

let couleur_fond = (rgb 205 205 205);


(*------------------------------------------------------------------------------*)

(* la classe des differents bateaux *)
class piece l = 
  val mutable lg = l (* longeur du bateaux *)
  val mutable nb_touche = create l false (* parties du bateaux detruites *)
  val mutable tab_position = create l {x=0;y=0} (* positions des pieces du bateau *)

  method enregistrer_position i a = tab_position.(i) <- a (* positionnement de chaque piece *)

  method reinit()=
    for i=0 to (lg-1) do
      nb_touche.(i) <- false;
      tab_position.(i)<- {x=0;y=0}
    done;



  method touche p =                                (* essai de toucher le bateau a la position p *)
    
    let i=ref 0 and t=ref false in
    while (!i<lg) & (not !t) do
      if (p=tab_position.(!i)) then 
        begin 
          nb_touche.(!i)<-true;t:=true
        end;
      incr i
    done;
    !t   (* retour si le bateau est touche *)

  method coule () = 
    let t = create lg true in       (* le bateau est-il coule *)
    t= nb_touche

end;;

type tplayer =
 < quiestu : unit -> string; touche : pos -> unit;
    positionnement : pos -> int -> int -> bool; placer_piece : string -> unit;
    attaque : pos -> coup;
    dessine_strategie : unit -> unit; dessine_plateau : string -> unit;
    partir : Join.location -> unit;
    jouer : tplayer * <<unit>> -> unit;
    perdu : unit -> bool;
    demande_saisie: int -> unit;
    demande_direction: unit -> int;
    mark: int -> int -> unit;
    unmark: int -> int -> unit;
    efface_msg : unit -> unit;
    rejouer : unit -> int;reinit : unit -> unit;
    affiche_msg : string -> color -> unit >


let not_initialized = ref true

(* La classe des joueurs *)
class player (name,t) as self =

  val nom = name 
  val type_piece = [|1;2;3;4|] (* type de bateaux disponible *)
  val pieces = init nbpiece (fun x->new piece t.(x))    (* Ses pieces *)
  val strategie = create_matrix n m (-1)  (* Les emplacements de ses pieces *)
  val plan = create_matrix n m Vide        (* Ses connaissances sur la strategie adverse *)
  val bateau = init 4 init_bateau
  val bateau_img = create 4 None
  val piece_bateau = init 7 init_piece_bateau
  val piece_bateau_img = create 7 None
  val case_H = 32 (* hauteur des cases *)
  val case_L = 32 (* largeur des cases *)
  val debut_tab_H = 30 (* on positionne le tableau en (debut_tab_H, debut_tab_L) *)
  val debut_tab_L = 30
  val inter_ligne = 50
  val ecran_L = 750
  val ecran_H = 600
      
      
  method reinit() =
    for i=0 to (n-1) do
      for j=0 to (m-1) do
        strategie.(i).(j)<-  -1;
        plan.(i).(j)<- Vide
      done;
    done;
    for i=0 to (nbpiece-1) do
      pieces.(i)#reinit()
    done;

(* ------ Quelques routines de gestion d'ecran -------- *)
             
 (* affichage d'un message m de couleur c *)
  method affiche_msg m c =
    moveto 10 420;
    set_font "lucidasans-bold-12";
    let f=foreground in
    set_color c;
    draw_string m;
    set_color f
 
  method efface_msg () =
    let f = foreground in
    set_color couleur_fond;
    fill_rect 6 403 410 185;
    set_color f;
  

  method quiestu () = nom
 
  (* On a touche un bateau adverse *)
  method touche p = plan.(p.x).(p.y) <- Touche
       
  method positionnement p l v =   (* Pour savoir si la piece est correctement placer dans le plan*)
                                  (* v = 0 : vertical *)
    let i = ref 0 
    and incx= 0 + v 
    and incy = 1 - v 
    and x=ref p.x 
    and y =ref p.y in
  
    try 
      while ((!i)<l) && (strategie.(!x).(!y) = (-1) ) do
        incr i;
        x:=!x+incx;y:=!y +incy;
      done;
      
      if !i>=l then
        begin
          x:=p.x; y:= p.y;
          
          (* bateau de longueur=1 *)
          if l=1 then strategie.(!x).(!y)<-0
          else
            begin
              (* l'arriere du bateau *)
              if v=0 then strategie.(!x).(!y) <- 6
              else strategie.(!x).(!y) <- 1;
              
              (* le milieu *)
              for i=1 to (l-2) do
                x:=!x+incx;y:=!y+incy;
                if v=0 then strategie.(!x).(!y) <-5 
                else strategie.(!x).(!y) <- 2;
              done;

              (* l'avant du bateau *)
              x:=!x+incx;y:=!y+incy;
              if v=0 then strategie.(!x).(!y) <- 4
              else strategie.(!x).(!y) <- 3;
            end;
          true;
        end
      else 
        false
    with _ -> false

  method demande_saisie t = 
    moveto 10 490;
    set_font "lucidasans-bold-12";
    draw_string("Placez la piece ");
    match bateau_img.(t-1) with
      Some(i) -> draw_image i 180 480;

            
  method demande_direction () = 
    self#affiche_msg "Quelle direction (v)erticale ou (h)orizontale:" black;
    let c=read_key() in
      if c='v' then 0 else 1;

  

  method mark a b =
    let f=foreground in
    set_color red;
    fill_circle (a*case_H+debut_tab_H+(case_H/2)) (b*case_L+debut_tab_L+(case_L/2)) 5;
    set_color f;

  method unmark a b =
    let f=foreground in
    set_color ( point_color(a*case_H+(debut_tab_H+4) )( b*case_L+(debut_tab_L+4) ) ) ;
    fill_circle (a*case_H+(debut_tab_H+10)) (b*case_L+(debut_tab_L+10)) 5; 
    set_color f;

    
  method placer_piece (nom)= (* Place les bateaux sur la grille du joueur*)
    self#dessine_plateau(nom);
    
    moveto 10 490;
    set_font "lucidasans-bold-12";
    draw_string("Patientez, s.v.p   ");
    (* precalcul des images *)
    if !not_initialized then
      begin
        for i=0 to (nbpiece-1) do
          bateau_img.(i)<-Some(make_image bateau.(i));
          draw_string "#"
        done;
        
        for i=0 to 6 do
          piece_bateau_img.(i)<-Some(make_image(piece_bateau.(i)));
          draw_string "#"
        done;
        not_initialized := false
      end;

    self#efface_msg();

      for i=0 to (nbpiece-1) do
     self#demande_saisie (type_piece.(i));
     
     let s= ref (wait_next_event [Button_down]) in 
     let a = ref ((!s.mouse_x-debut_tab_H) / case_H) and b = ref ((!s.mouse_y-debut_tab_L)/case_L) in
     self#mark (!a) (!b);
     let vert = ref (self#demande_direction()) in
     
     while not(self#positionnement {x=(!a);y=(!b)} type_piece.(i) (!vert) ) do
       self#unmark (!a) (!b);
       s:= wait_next_event [Button_down];
       a:= ((!s.mouse_x-debut_tab_H) / case_H);
       b:= ((!s.mouse_y-debut_tab_L) / case_L);
       self#mark (!a) (!b);
       vert:= self#demande_direction() ; 
     done;

     
     let incx= 0 + !vert 
     and incy = 1 - !vert in
     
     for j=0 to type_piece.(i)-1 do
       pieces.(i)#enregistrer_position j {x=(!a);y=(!b)};
       a:=!a+incx;b:=!b+incy;
     done;
     
     self#dessine_strategie();
     self#efface_msg();
   done

  method attaque p = (* attaque a la position p, reponse: Loupe, Touche, Coule ou Vide*)
 
      let i=ref 0 in
      while (!i<nbpiece) && (not (pieces.(!i)#touche(p)) ) do
        incr i;
      done;
      if !i>=nbpiece then 
        begin
          draw_image (make_image eau) (debut_tab_H+ p.x*case_H+1 ) (debut_tab_L+ p.y*case_L+1 );
          Loupe
        end
      else 
        if pieces.(!i)#coule() then
          begin
            draw_image (make_image coule) (debut_tab_H+ p.x*case_H+1 ) (debut_tab_L+ p.y*case_L+1 );
            Coule 
          end
        else 
          begin
            draw_image (make_image boom) (debut_tab_H+ p.x*case_H+1 ) (debut_tab_L+ p.y*case_L+1 );
            Touche;
          end
 
  method dessine_strategie () = (* dessine les pieces du joueur *)
    let f=foreground in
    set_color red;
    for i=0 to (n-1) do
      for j=0 to (m-1) do
        if (strategie.(i).(j)<>(-1)) then 
          match piece_bateau_img.(strategie.(i).(j)) with
            Some(x)-> draw_image(x) (debut_tab_H+i*case_H+1) (debut_tab_L+j*case_L+1) ;
      done;
    done;
    set_color f;   

    
  method dessine_plateau nom_adversaire = (* dessin du plateau de depart *)
    open_graph (" "^string_of_int(ecran_L)^"x"^string_of_int(ecran_H));
    
    let f=foreground in
    set_color couleur_fond;
    fill_rect 0 0 ecran_L ecran_H;
    
    (* se'paration *)
    set_line_width 2;
    
    moveto 5 5;
    set_color white;
    lineto 5 380;
    lineto (ecran_L/2) 380;
    set_color black;
    lineto (ecran_L/2) 5;
    lineto 5 5;

    
    moveto (ecran_L/2+10) 5;
    set_color white;
    lineto (ecran_L/2+10) 380;
    lineto (ecran_L-5) 380;
    set_color black;
    lineto (ecran_L-5) 5;
    lineto (ecran_L/2+10) 5;

    moveto 5 400;
    set_color white;
    lineto 5 590;
    lineto (ecran_L-5) 590;
    set_color black;
    lineto (ecran_L-5) 400;
    lineto 5 400;

    set_color f;

   (* Premier cadran*)
    let c1 = (case_H*n) and c2 = (case_L*m) in 
    let img = (make_image fond_ecran) in
    for i=0 to 9 do
      for j=0 to 9 do
       draw_image img (debut_tab_H+i*case_H+1) (debut_tab_L+j*case_L+1);
      done;
    done;

    moveto debut_tab_H debut_tab_L;
    lineto (debut_tab_H+c1) debut_tab_L;
    lineto (debut_tab_H+c1) (debut_tab_L+c2);
    lineto debut_tab_H (debut_tab_L+c2);
    lineto debut_tab_H debut_tab_L;

   (* deuxieme cadran *)
    let c3 = (case_L*n)+inter_ligne in 
    for i=0 to 9 do
      for j=0 to 9 do
       draw_image img (debut_tab_H+c3+i*case_H+1) (debut_tab_L+j*case_L+1);
      done;
    done; 

    moveto (debut_tab_H+c3) debut_tab_L;
    lineto (debut_tab_H+c3+c1) debut_tab_L;
    lineto (debut_tab_H+c3+c1) (debut_tab_L+c2);
    lineto (debut_tab_H+c3) (debut_tab_L+c2);
    lineto (debut_tab_H+c3) debut_tab_L;

   (* cadrillage des cadrans *) 
    set_line_width 1;
    set_font "lucidasans-bold-8";
    for i=1 to n do
      moveto (debut_tab_H+i*case_H) debut_tab_L;
      lineto (debut_tab_H+i*case_H) (debut_tab_L+c2);
      
      moveto (debut_tab_H+c3+i*case_H) debut_tab_L;
      lineto (debut_tab_H+c3+i*case_H) (debut_tab_L+c2);
      
      moveto (debut_tab_H+(i-1)*case_H+8) (debut_tab_L-15);
      draw_string (string_of_int(i-1));
      moveto (debut_tab_H+c3+(i-1)*case_H+8) (debut_tab_L-15);
      draw_string (string_of_int(i-1));
    
    done; 

    for i=1 to m do
      moveto debut_tab_H (debut_tab_L+i*case_L);
      lineto (debut_tab_H+c1) (debut_tab_L+i*case_L);
      
      moveto (debut_tab_H+c3) (debut_tab_L+i*case_L);
      lineto (debut_tab_H+c3+c1) (debut_tab_L+i*case_L);
      
      moveto (debut_tab_H-10) (debut_tab_L+(i-1)*case_L);
      draw_string (string_of_int(i-1));
      moveto (debut_tab_H+c3-10) (debut_tab_L+(i-1)*case_L);
      draw_string (string_of_int(i-1));
    done;

    moveto 500 520;
    set_font "lucidasans-bolditalic-18";
    draw_string "Bataille Navale";

    moveto 150 360;
    set_font "lucidasans-bold-12";
    draw_string (self#quiestu());

    moveto 550 360;
    set_font "lucidasans-bold-12";
    draw_string nom_adversaire;
    
  method partir l = Join.go(l)
      
  method jouer ((p : tplayer),replay) =
    let def j!() = 
      {
      (
      moveto 10 420;
      let f = foreground in
      set_color red;
      draw_string("Quelle position ? ");
      set_color f;
      let s= ref (wait_next_event [Button_down]) in 
      let a = ref ((!s.mouse_x-(n*case_H+inter_ligne+debut_tab_H)) / case_H) and b = ref ((!s.mouse_y-debut_tab_L)/case_L) in      
      let c = ref (p#attaque {x=(!a);y=(!b)}) in
           
      while (!a>9) || (!a<0) || (!b>9) || (!b<0) do
        s:= wait_next_event [Button_down] ; 
        a := ((!s.mouse_x-(n*case_H+inter_ligne+debut_tab_H)) / case_H) ;
        b := ((!s.mouse_y-debut_tab_L)/case_L);
        c := p#attaque {x=(!a);y=(!b)};
      done;

      self#efface_msg();

      match (!c) with
        Loupe -> 
          plan.(!a).(!b)<-Loupe;
          let c3 = (case_H*n)+inter_ligne in
          draw_image (make_image eau) (debut_tab_H+c3+(!a)*case_H+1) (debut_tab_L+(!b)*case_L+1);
          (* sound 440 100;*)
          p#jouer ( (self :> tplayer), replay)
      | _ -> 
          plan.(!a).(!b)<-Touche;
          let c3 = (case_H*n)+inter_ligne in
          
          draw_image (make_image boom) (debut_tab_H+c3+(!a)*case_H+1) (debut_tab_L+(!b)*case_L+1);
          match (!c) with
            Coule ->
              if p#perdu() then
                begin
                  moveto 500 190;
              (*   
                  set_font "-adobe-zapfdingbats-*-*-normal-*-30-*-*-*-*-*-*-*";
                  draw_string "O";*)
                  set_color red;
                  set_font "lucidasans-bolditalic-24"; 
                  set_text_size 30;
                  draw_string("GAGNE");
                  (*set_font "-adobe-zapfdingbats-*-*-normal-*-30-*-*-*-*-*-*-*";
                  draw_string "O";*)
                  if (self#rejouer())=1 then 
                    if (p#rejouer())=1 then 
                      begin
                        self#reinit();
                        p#reinit();
                        {| replay() }
                      end
                    else self#affiche_msg ("Desole votre partenaire ne veux plus jouer... au revoir") red
                end
              else
                p#jouer ( (self :> tplayer), replay)
          | 
            Touche ->
              p#jouer ( (self :> tplayer),replay) );} in
    {| j() }
      
  method perdu () = 
    let i = ref 0 in
    while (!i<nbpiece) && (pieces.(!i)#coule()) do 
      incr i; 
    done;
    if !i=nbpiece then 
      begin
        moveto 150 190;
        set_color (rgb 255 0 0);
        set_font "lucidasans-bolditalic-24"; 
        set_text_size 30;
       
        draw_string "PERDU";
        true
      end
      else false

  method rejouer() = 
    moveto 10 420;
    set_font "lucidasans-bold-12";
    draw_string("Voulez-vous rejouer? (o/n)");
    (* sound 440 100;*)
    let c=read_key() in
    self#efface_msg();
    if c='o' then 1 else 0;

end;;
    

let def bataille_navale(l,name,_) | pas_de_joueur!() =
   {attente_partenaire(l,name) | reply } 
or bataille_navale (l1,nom1,_) | attente_partenaire!(l2,nom2) = 
  reply | 
  let p1 = new player (nom1,[|1;2;3;4|]) 
  and p2 = new player (nom2,[|1;2;3;4|]) in
  let def rejouer!() = 
    {
     (p1#partir(l1));  
     (p1#placer_piece nom2);
     (p1#jouer (p2,rejouer));}      
   |
    {
     (p2#partir(l2));  
     (p2#placer_piece nom1);}
   | 
     pas_de_joueur() in
  
  { rejouer() };;

{| pas_de_joueur()};;

let _ = 
  Arg.parse (Genserver.ns_args ()) (fun _ -> ()) "start Navy-war server";
  Genserver.server bataille_navale (vartype: (Join.location * string * string list -> unit)
    metatype)




