(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Xlib;;
open Xtypes
open Xconst
open Xutils 

let dpy = Xlib.openDisplay (Sys.getenv "DISPLAY")
let root = defaultRoot dpy;;
let black= defaultBlack dpy;;
let white= defaultWhite dpy;;
print_string "_";print_newline ();;
let win = X.createWindow dpy root 0 0 200 200 copyDepthFromParent InputOutput 
            copyVisualFromParent 2 [CWBackPixel black];;
print_string "_";print_newline ();;
sync dpy;;

print_string "_";print_newline ();;

let fixed = X.openFont dpy "fixed";;
let gc = X.createGC dpy root [GCforeground white;
                               GCbackground black;
                               GCfont fixed;
                               GCfonction GXcopy];;
sync dpy;;

print_string "_";print_newline ();;


selectInput dpy win [ExposureMask;StructureNotifyMask;KeyPressMask;
                    EnterWindowMask;LeaveWindowMask];;

X.mapWindow dpy win;;
sync dpy;;

exception AbortMainLoop;;
exception MismatchEvent;;

print_string "+";print_newline ();;


let r = { x=0; y=0; dx=200; dy=200 }
in
try
while true do
  let e = nextEvent dpy 
  in
    match e.ev_event with
        KeyPressEvent k -> exit 0
      | ButtonReleaseEvent k -> exit 0
      | ExposeEvent e ->
          for i=2 to e.Xexpose.count do
            let {ev_type=t} = nextEvent dpy in
              if t<> Expose then raise MismatchEvent
          done;
          clearWindow dpy win;
          Xlib.drawString dpy win gc 50 50 "bonjour1" 0 8;
          X.polyArc dpy win gc [(r.x,r.y,r.dx,r.dy,1,(360*64))]
      | ConfigureNotifyEvent b ->
          r.x <- b.Xconfigure.x;
          r.y <- b.Xconfigure.y;
          r.dx <- b.Xconfigure.width;
          r.dy <- b.Xconfigure.height
      | _ -> print_int  e.ev_num; print_newline ()
done;
NoError
with
    XError e -> Error e
;;
print_string "+";print_newline ();;

Xlib.nextEvent dpy;;
print_string "+";print_newline ();;

X.destroyWindow dpy win;;


