(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Efuns
open Text
open Frame
open Simple
open Select
open Keymap
  
let create_bindings location =
  let actives = get_local_var (Location location) interactives in
  let s = ref "Default bindings:" in
  List.iter (fun (name,(_,binding)) ->
    match !binding with
      None -> ()
    | Some key_list ->
        s := Printf.sprintf "%s\n%s : %s" !s 
             (Keymap.print_key_list (List.rev key_list)) name
             ) actives;
  let text = Text.create !s in
  Buffer.create location "*bindings*" text (Keymap.create ())

let meta_hist = ref []
let call_interactive frame =
  let buf = frame.frm_buffer in
  let buf_actives = get_local_var (Buffer buf) interactives in
  let loc_actives = 
    get_local_var (Location buf.buf_location) interactives in
  let actives = buf_actives @ loc_actives in
  select frame "M-x " meta_hist ""
    (fun _ -> List.map fst actives)
    (fun s -> s)
    (fun name -> 
      try
        let f, key = List.assoc name actives in
        begin
          match !key with
            None -> ()
          | Some key_list ->
              let top_window = Window.top frame.frm_window in
              Top_window.message top_window
                ("you can run "^name^" by typing "^
                 (Keymap.print_key_list (List.rev key_list)))
        end;
        f frame
      with
        Not_found -> Obj.repr 0)

