(***********************************************************************)
(*                                                                     *)
(*                               Efuns                                 *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* Reimplement regexps with marshallable regexps *)

open Str

type regexp
  
external compile_regexp: string -> bool -> regexp = "str_compile_regexp"

let regexp_string = "regexp"

let marshal_regexp o =
  (Obj.field o 3,Obj.field o 4)

let create str fold =
  let regexp = compile_regexp str fold in
  let o = Obj.new_block Tag.marshal 5 in
  Obj.set_field o 0 (Obj.repr regexp_string);
  Obj.set_field o 1 (Obj.repr marshal_regexp);
  Obj.set_field o 2 (Obj.repr regexp);
  Obj.set_field o 3 (Obj.repr str);
  Obj.set_field o 4 (Obj.repr fold);
  (Obj.magic o: regexp)
  
let unmarshal_regexp (str,fold) = create str fold

let _ = Junmarshal.new_unmarshaler regexp_string unmarshal_regexp  
  
let repr (regexp: regexp) = 
  (Obj.magic (Obj.field (Obj.repr regexp) 2): Str.regexp)

let compile_regexp str fold = create str fold
  
let quote = quote

let regexp e = compile_regexp e false

let regexp_case_fold e = compile_regexp e true

let regexp_string s = compile_regexp (quote s) false

let regexp_string_case_fold s = compile_regexp (quote s) true

let string_match regexp string int = string_match (repr regexp) string int

let search_forward regexp str int = search_forward (repr regexp) str int
let search_backward regexp str int =  search_backward (repr regexp) str int
let string_partial_match regexp str int = 
  string_partial_match (repr regexp) str int
let matched_string = matched_string
let match_beginning = match_beginning
let match_end = match_end
let matched_group = matched_group
let group_beginning = group_beginning
let group_end = group_end
let global_replace regexp str1 str2 = global_replace (repr regexp) str1 str2
let replace_first regexp str1 str2 = replace_first (repr regexp) str1 str2
let global_substitute regexp f str = global_substitute (repr regexp) f str
let substitute_first regexp f str = substitute_first (repr regexp) f str 
let split regexp string = split (repr regexp) string
let bounded_split regexp string int = bounded_split (repr regexp) string int
let string_before = string_before
let string_after= string_after
let first_chars = first_chars
let last_chars = last_chars



external replacement_text: string -> string -> string = "str_replacement_text"

let replace_matched repl matched =
  replacement_text repl matched
  
