(* A sample file to see how to dynamically configure the system:
Efuns.add_start_hook is executed after the file is loaded. As a consequence, 
  we can change the function "fill_paragraph" for buffer "main.tex".
*)

open  Efuns
open Text
open Frame
open Top_window
open Simple
open Interactive
open Keymap

  
  
let simplify text start point =
  let start = dup_point text start in
  let rec iter last_c =
    if start < point then
      let c = get_char text start in
      if c = ' ' || c = '\n' || c = '\t' then
        ( delete text start 1; 
          iter ' ')
      else
      if last_c = ' ' then
        ( insert text start " ";
          fmove text start 2;
          iter 'a')
      else
        ( fmove text start 1;
          iter 'a')
  in
  iter 'a';
  remove_point text start

let fill_paragraph frame =
  let buf = frame.frm_buffer in
  let text = buf.buf_text in
  let point = frame.frm_point in
  let start = dup_point text point in
  let session = start_session text in
  let syntax = buf.buf_syntax_table in
  backward_paragraph buf start;
  let fin = dup_point text start in
  forward_paragraph buf fin;
  simplify text start fin;
  insert text start "\n";
  
  forward_word buf start;
  while (forward_word buf start; start < fin) do
    if point_to_bol text start > 75 then
      ( backward_word buf start;
        ( if bmove text start 1 = 1 then
            let c = get_char text start in
            if c = ' ' then
              ( delete text start 1;
                insert text start "\n";
                ())
            else
              ( fmove text start 1;
                insert text start "\n";
                ())
        );
        forward_word buf start)
  done;
  insert text fin "\n";
  remove_point text start;
  remove_point text fin;
  commit_session text session
  
  
let _ =  
  Efuns.add_start_hook (fun location ->
      print_string "Init"; print_newline (); 
      let buf = Buffer.default location "main.tex" in
      interactive (Buffer buf) [MetaMap, Char.code 'q']
        "fill-paragraph" (fill_paragraph);
      buf.buf_syntax_table.(Char.code '\\') <- true;
      buf.buf_syntax_table.(Char.code '\'') <- true;
      buf.buf_syntax_table.(Char.code '-') <- true;

 
  )

