(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Xtypes
open Concur

let allocId = Display.alloc_id
let openDisplay = Display.openDisplay
let closeDisplay = Display.closeDisplay

(*[X.nextEvent] nextEvent dpy :
 on renvoie le dernier evenement *)
let nextEvent dpy =
  Mutex.lock dpy.dpy_mutex;
  let rec iter () =
    while Xfifo.empty dpy.event_queue do
      Condition.wait dpy.dpy_wait dpy.dpy_mutex
    done;
    let ev = Xfifo.take dpy.event_queue in
    if ev.ev_consumed then iter ()
    else ev
  in
  let ev = iter () in
  ev.ev_consumed <- true;
  Mutex.unlock dpy.dpy_mutex;
  ev

(*[X.checkEvent] checkEvent dpy :
 on lit le dernier evenement sans le sortir de la queue *)
let checkEvent dpy =
  try
    Xfifo.read dpy.event_queue
  with
    _ -> raise Not_found

(*[X.peekEvent] peekEvent dpy :
 on lit le dernier evenement sans le sortir de la queue *)
let peekEvent dpy =
  Mutex.lock dpy.dpy_mutex;
  while Xfifo.empty dpy.event_queue do
    Condition.wait dpy.dpy_wait dpy.dpy_mutex
  done;
  let ev = Xfifo.read dpy.event_queue in
  Mutex.unlock dpy.dpy_mutex;
  ev

(* Prevent nextEvent from reading this event *)
let consumeEvent ev = ev.ev_consumed <- true

let createSimpleWindow dpy parent x y width height border_width args =
  let screen = dpy.dpy_roots.(dpy.dpy_screen_default) in
  X.createWindow dpy parent x y width height 
    screen.scr_root_depth InputOutput 
    screen.scr_root_visual_id  border_width args
    

let drawString dpy win gc x y str pos len =
  let len = min len (String.length str - pos) in
  X.polyText8 dpy win gc x y [Text8 (0,str,pos,len)]

let imageString dpy win gc x y str pos len =
  let len = min len (String.length str - pos) in
  X.imageSubText8 dpy win gc x y str pos len

let printXError error = Display.print_xerror error

let moveWindow dpy win x y =
  X.configureWindow dpy win [CWX x; CWY y]

let resizeWindow dpy win dx dy =
  X.configureWindow dpy win [CWWidth dx; CWHeight dy]
