/***********************************************************************/
/*                                                                     */
/*                           NPRremote                                 */
/*                                                                     */
/*          Francois Rouaix, projet Cristal, INRIA Rocquencourt        */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include "npapi.h"

#include "npremote.h"

#include <signal.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>


typedef struct _PluginInstance
{
  int instance_id;		/* number chosen by the server */
  Display *display;		/* as given by SetWindow */
  Window window;		/* as given by SetWindow */
} PluginInstance;


/* Adress of server */
struct sockaddr_un serveraddr;

/* file descriptors are valid only if inited is true */
int inited = 0;

#define checkinit() if (!inited) return NPERR_GENERIC_ERROR


/* master connection for transmitting
    New, Destroy, SetWindow, Print
 */
int to_server = -1;
int x =0;
/* master connection for receiving
    UserAgent, Status, GetURL
 */
int from_server = -1;

/* installed callback for NPN calls */
XtInputId cbid = 0;




char*
NPP_GetMIMEDescription(void)
{
  return(NPRem_MIME);
}

NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
  NPError err = NPERR_NO_ERROR;
  
  switch (variable) {
  case NPPVpluginNameString:
    *((char **)value) = NPRem_Name;
  break;
  case NPPVpluginDescriptionString:
    *((char **)value) = NPRem_Desc;
  break;
  default:
    err = NPERR_GENERIC_ERROR;
  }
  return err;
}

jref
NPP_GetJavaClass()
{
	return NULL;
}

#define CMD_INIT  "INIT"
#define CMD_CALLB "CALLB"
#define CMD_NEW   "NEW"
#define CMD_KILL  "KILL"
#define CMD_STREAM "STREAM"
#define CMD_PRINT  "PRINT"
#define CMD_NOWIN  "NOWIN"
#define CMD_WIN    "WIN"

NPError
NPP_Initialize(void)
{
  if (!inited) {
    static char buf[1024];
    /* Start the viewer server */
    sprintf(buf,NPRem_Command,getpid());
    if (system(buf)) {
      perror("system");
      return NPERR_GENERIC_ERROR;
    }
    /* Built the server address */
    serveraddr.sun_family = AF_UNIX;
    sprintf(buf,NPRem_Socket,getenv("HOME"),getpid());
    bcopy (buf,serveraddr.sun_path, strlen(buf));
    
    /* create the two main connections */
    if (-1 == (to_server = socket(AF_UNIX, SOCK_STREAM, 0))) {
      perror("socket");
      return NPERR_GENERIC_ERROR;
    };
    
    if (-1 == (from_server = socket(AF_UNIX, SOCK_STREAM, 0))) {
      perror("socket");
      close(to_server);
      return NPERR_GENERIC_ERROR;
    };
    
    if (connect (to_server, (struct sockaddr *)&serveraddr, 
		 sizeof (serveraddr)) == -1) {
      perror("connect");
      fprintf(stderr, "Can't connect to %s\n", buf);
      close(to_server);
      close(from_server);
      return NPERR_GENERIC_ERROR;
    };
    writestring(to_server, CMD_INIT, strlen(CMD_INIT));
    
    /*
    if (connect (from_server, (struct sockaddr *)&serveraddr, 
		 sizeof (serveraddr)) == -1) {
      perror("connect");
      fprintf(stderr, "Can't connect to %s\n", buf);
      close(to_server);
      close(from_server);
      return NPERR_GENERIC_ERROR;
    };
    */
/*    writestring(from_server, CMD_CALLB, strlen(CMD_CALLB));*/

    /* We're connected. Inform the server about the master cnx */
    fprintf(stderr, "Started\n");
    inited = 1;
    return NPERR_NO_ERROR;
  }
  else { /* already initialised ! */
    return NPERR_GENERIC_ERROR;
  }
}

void
NPP_Shutdown(void)
{    static char buf[1024];

  if (cbid != 0) {
    XtRemoveInput(cbid);
    cbid = 0;
  };
  /* this supposedly closes the server */
  close(to_server);
  close(from_server);
  inited = 0;
  sprintf(buf,NPRem_Socket,getenv("HOME"),getpid());
  unlink (buf);
/*  fprintf (stderr,"on quitte le plg\n");
  fflush (stderr);*/
  return;
}

/* A list of instances : we need that for callbacks from the server */
struct InstanceCell {
  NPP car;
  struct InstanceCell *cdr;
};

struct InstanceCell *allinstances = NULL;

void
addinstance(NPP inst) 
{
  struct InstanceCell *cell;
  if (NULL == (cell = NPN_MemAlloc(sizeof(struct InstanceCell *))))
    return;
  cell->car = inst;
  cell->cdr = allinstances;
  allinstances = cell;
  return;
}

void
reminstance(int n)
{
  struct InstanceCell *ptr, *last;

  for (ptr=allinstances; ptr != NULL; ptr = ptr->cdr) {
    PluginInstance *This = (PluginInstance *)ptr->car->pdata;
    if (This->instance_id == n) {
      if (ptr == allinstances) allinstances = ptr->cdr;
      else last->cdr = ptr->cdr;
      NPN_MemFree(ptr);
      break;
    };
    last = ptr;
  }
}

NPP
getinstance(int n)
{
  struct InstanceCell *ptr;

  for (ptr=allinstances; ptr != NULL; ptr = ptr->cdr) {
    PluginInstance *This = (PluginInstance *)ptr->car->pdata;
    if (This->instance_id == n) return(ptr->car);
  }
  return NULL;
}
  


/* too bad we don't get the "referer" here */
NPError 
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode,
	int16 argc,
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{

/*     DisplayJavaMessage(instance,"Ici, on commence NEW",-1); */
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;

  checkinit();
  
  instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
  if (instance->pdata == NULL) {
    return NPERR_OUT_OF_MEMORY_ERROR;
  }
  else {
    PluginInstance* This = instance->pdata;
    int i;

    /* Write the instance request */
    /* TODO: look at the mode stuff */ 
    writestring(to_server, CMD_NEW, strlen(CMD_NEW));
  /*  writestring(to_server, pluginType, strlen(pluginType));*/
    writeint(to_server, argc);
    for (i=0;i<argc;i++) {
      writestring(to_server,argn[i], strlen(argn[i]));
      writestring(to_server,argv[i], strlen(argv[i]));
    };

    /* we get back an id */
    This->instance_id = readint(to_server);
    This->window = (Window)NULL;
    if (This->instance_id == -1)
      return NPERR_GENERIC_ERROR;
    else {
      addinstance(instance);
      return NPERR_NO_ERROR;
    }
  }
/*     DisplayJavaMessage(instance,"Ici, on fini NEW",-1); */
}

NPError 
NPP_Destroy(NPP instance, NPSavedData** save)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  checkinit();

  This = (PluginInstance*) instance->pdata;
  
  if (This != NULL) {
    writestring(to_server, CMD_KILL, strlen(CMD_KILL));
    writeint(to_server, This->instance_id);
    readack(to_server);
    reminstance(This->instance_id);
    NPN_MemFree(instance->pdata);
    instance->pdata = NULL;
  }
  
  return NPERR_NO_ERROR;
}


/* Stream stuff */
typedef struct _StreamData
{
  int socket;
} StreamData;



NPError 
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream *stream, 
	      NPBool seekable,
	      uint16 *stype)
{

  StreamData *sdata;
  PluginInstance *This;
/*     DisplayJavaMessage(instance,"Ici, NEWSTREAM",-1); */
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;

  This = (PluginInstance *)instance->pdata;

  if (stream == NULL)
    return NPERR_GENERIC_ERROR;

  checkinit();

  /* Allocate my information for the stream */
  sdata = (StreamData *)NPN_MemAlloc(sizeof(StreamData));
  if (NULL == sdata) return NPERR_OUT_OF_MEMORY_ERROR;
  stream->pdata = sdata;

  /* Connect to the server */
  if (-1 == (sdata->socket = socket(AF_UNIX, SOCK_STREAM, 0))) {
    perror("socket");
    NPN_MemFree(sdata);
    stream->pdata = NULL;
    return NPERR_GENERIC_ERROR;
  };

  if (connect (sdata->socket, (struct sockaddr *)&serveraddr, 
	       sizeof (serveraddr)) == -1) {
    perror("connect");
    fprintf(stderr, NPRem_What " server is not responding.\n");
    close(sdata->socket);
    NPN_MemFree(sdata);
    stream->pdata = NULL;
    return NPERR_GENERIC_ERROR;
  };

  /* Write a bunch of information about the stream */
/*  writestring(sdata->socket, CMD_STREAM, strlen(CMD_STREAM));
  writeint(sdata->socket, This->instance_id);
  writestring(sdata->socket, type, strlen(type)); */
  /* also write information contained in stream */
  /* TODO: we should check url is non-null */
/*  writestring(sdata->socket, stream->url, strlen(stream->url));
  writeint(sdata->socket, stream->end);
  writeint(sdata->socket, stream->lastmodified);
  readack(sdata->socket);*/
  return NPERR_NO_ERROR;
}

int32 
NPP_WriteReady(NPP instance, NPStream *stream)
{
  if (stream != NULL) {
    StreamData *sdata = stream->pdata;
    fd_set wfds;
    struct timeval tv;

    if (sdata == NULL)
      return 0;
	
    FD_ZERO(&wfds);
    FD_SET(sdata->socket, &wfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    if (select(sdata->socket+1,NULL,&wfds,NULL,&tv))
      return PIPE_BUF;		/* ???? */
    else return 0;
  }
  else
    return 0;
}

int32 
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
  if (stream != NULL) {
    StreamData *sdata = (StreamData *)stream->pdata;
    return(write(sdata->socket,buffer,len));
  }
  else return 0;
}

NPError 
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
  if (stream != NULL) {
    StreamData *sdata = stream->pdata;
    close(sdata->socket);
    NPN_MemFree(sdata);
    stream->pdata = NULL;
    return NPERR_NO_ERROR;
  }
  else
    return NPERR_INVALID_INSTANCE_ERROR;
}

void 
NPP_Print(NPP instance, NPPrint* printInfo)
{
  if(printInfo == NULL)
    return;
  
  checkinit();

  if (instance != NULL) {
    PluginInstance* This = (PluginInstance*) instance->pdata;
    writestring(to_server, CMD_PRINT, strlen(CMD_PRINT));
    writeint(to_server,This->instance_id);
    readack(to_server);
  }
}

/* The other way around : the plugin acts as a server for the Server.
   This function is invoked when Server connects to CSock because
   we put an Xt file hander on it.
   BUG: When error happens, Server doesn't get the expected answers
 */
/* NPN_GetURL without target (will cause NewStream to happen) */
#define CB_GETURL  "GETURL"
/* NPN_GetURL with target */
#define CB_DISPURL "DISPURL"
/* NPN_Status */
#define CB_MSG     "MSG"
/* NPN_UserAgent */
#define CB_AGENT   "AGENT"
/* TODO: complete with other Netscape Methods */

void callback(XtPointer clientdata, int *fd, XtInputId *id)
{
  char *cmd;
  int n;
  NPP inst = NULL;
  cmd = readstring(*fd);	/* the command */
  if (cmd == NULL) { /* EOF most probably */
    XtRemoveInput(*id);
    cbid = 0;
    return;
  };

  n = readint(*fd);		/* the originating instance id */
  if (n == -1) {
    fprintf(stderr,"Wrong NPRemote protocol. Can't read instance number\n");
    goto bailout;
  }

  inst = getinstance(n);	/* ok, get the instance */
  if (inst == NULL) {
    fprintf(stderr,"Invalid NPRemote instance %d\n",n);
    goto bailout;
  }

  /* debugging   fprintf(stderr, "Got %s %d\n", cmd, n); */

  /* 
   * Dispatch the command 
   */

  /* Get an URL for the plugin */
  if (0 == strncmp(CB_GETURL, cmd, strlen(CB_GETURL))) {
    char *url = readstring(*fd);

    if (url != NULL) {
      NPN_GetURL(inst, url, NULL);
      NPN_MemFree(url);
    };
    /* NN will call us back through NewStream. Nothing more to do here. */
    writeack(*fd);
  }

  /* Display an URL in Netscape */
  else if (0 == strncmp(CB_DISPURL, cmd, strlen(CB_DISPURL))) {
    char *url = readstring(*fd);
    char *target = readstring(*fd);

    if ((url != NULL) && (target != NULL))
      NPN_GetURL(inst, url, target);
    if (url != NULL)  NPN_MemFree(url);
    if (target != NULL) NPN_MemFree(target);
    writeack(*fd);
  }

  /* status message */
  else if (0 == strncmp(CB_MSG, cmd, strlen(CB_MSG))) { 
    char *msg = readstring(*fd);
    
    if (msg != NULL) {
      NPN_Status(inst, msg);
      NPN_MemFree(msg);
    }
    writeack(*fd);
  }

  /* user agent */
  else if (0 == strncmp(CB_AGENT, cmd, strlen(CB_AGENT))) {
    const char *s = NPN_UserAgent(inst);
    writestring(from_server, s, strlen(s));
  }

bailout:
  NPN_MemFree(cmd);
  return;
}


/* This creates an file handler on CSock */
void extrainit(PluginInstance *this) {
  static inited = 0;
  if (!inited) {
    Widget netscape_widget;
    XtAppContext appcon;
    netscape_widget = XtWindowToWidget(this->display, this->window);
    appcon = XtWidgetToApplicationContext(netscape_widget);
    if (appcon != NULL) {
      cbid = XtAppAddInput(appcon, from_server,(XtPointer) XtInputReadMask,
			   callback, NULL);
       inited = 1;
    }
  }
}



NPError 
NPP_SetWindow(NPP instance, NPWindow* window)
{
  PluginInstance* This;
      int awin;
    Window child;

/*     DisplayJavaMessage(instance,"Ici, SETWIN",-1); */
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  This = (PluginInstance*) instance->pdata;

  checkinit();

  /* If either window or window->window are NULL, the plug-in must not
     perform any additional graphics operations on the window and should
     free any resources associated with the window.  
  */
  
if ((window == NULL) || (window->window == NULL)) {
    This->window = (Window)NULL;
    return NPERR_NO_ERROR;
  }
  else { 

    /*fprintf(stderr,"Je suis SetWin : %d\n",This->window);    */
    This->window = (Window) window->window;
    This->display = ((NPSetWindowCallbackStruct *)window->ws_info)->display;

    /* extrainit(This);  */
    writestring(to_server, CMD_WIN, strlen(CMD_WIN));
    writeint(to_server,This->instance_id);
    writeint(to_server,window->x);
    writeint(to_server,window->y);
    writeint(to_server,window->width);
    writeint(to_server,window->height);
    /* In response, we get the window id of the window */
    XSync(This->display, False); 
    child = readint(to_server);
    XUnmapWindow(This->display,child);
    XFlush(This->display); 

    XReparentWindow(This->display,child,This->window,0,0); 
/*    XFlush(This->display); */
    XMapWindow(This->display,child);
    XMapWindow(This->display,This->window);

    XFlush(This->display); 

    return NPERR_NO_ERROR;
  }

}

/* These are not yet implemented */
void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
fprintf(stderr,"StreamAsFile not implemented in NPRemote plugin\n");
return;
}

void
NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
fprintf(stderr,"URLNotify not implemented in NPRemote plugin\n");
return;
}








