(***********************************************************************)
(*                                                                     *)
(*                           Pong for JoCaml                           *)
(*                                                                     *)
(*            Sylvain Conchon, projet Para, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)
(* Programme ping-pong *)

open Graphics;;
open Array;;



let offset_H = 40
let ecran_L = 500 and ecran_H = 380;;


let barre_L = 40 and barre_E = 6;; (*hauteur et epaisseur des barres*)
(*longueur et hauteur du terrain*)
let terrain_L = 500 and terrain_H = ecran_H - 2 * offset_H;; 
let m_incr = 4;;
(* min_x, max_x : position des barres des joueurs p1 et p2*)
let min_x = 40 and max_x = (terrain_L-barre_E-40) and dep_y = (terrain_H/2);;
let mil_x = terrain_L/2 and mil_y = terrain_H /2 ;;
let etat_y = terrain_H + offset_H + 20
let adv_x = ecran_L - 200
let min_pos = offset_H + (barre_L / 2)
let max_pos = offset_H + terrain_H - (barre_L / 2)

let mouse_pos () = 
  let y = snd (mouse_pos ()) in
  let y = if y < min_pos then min_pos
    else
    if y > max_pos then max_pos 
    else y
  in
  y - min_pos

class balle (px,py) = 
  val mutable posx = px
  val mutable posy = py
  val mutable pposx = px
  val mutable incx = 4
  val mutable incy = 4
  val mutable valeur = 4
  val mutable cpt = 0
  val mutable pointp1 = 0
  val mutable pointp2 = 0
 

  method partir(l) = go (l) 
  
  method jouer(putpos,get_p1,get_p2,(beep:unit->unit)) =
    while true do
      let xp1,yp1 = get_p1() and
        xp2,yp2 = get_p2() in
      
      
      if posy>terrain_H then incy <- -valeur;
      if posy<0 then incy <- valeur;
      
      if posx>terrain_L then incx <- -valeur;
      if posx<0 then incx<- valeur;
      
      (*pposy<posy : retour de la balle pour le joueur P1*)
     
      
      if (pposx>=posx)&&(posx>=xp1)&&(posx<=xp1+barre_E)
          &&(posy>yp1)&&(posy<yp1+barre_L) then 
        begin
          cpt <- (cpt +1) mod 6;
          valeur <- valeur +(cpt/5);
          incx <- valeur;
          incy <- if incy>0 then valeur else -valeur;
          pointp1 <- pointp1 + 1;
          beep();
          if pointp1 = 21 then pointp1 <- 0
        end;

      if  (pposx<=posx)&&(posx>=xp2)&&(posx<=xp2+barre_E)
          &&(posy>yp2)&&(posy<yp2+barre_L) then
        begin
          cpt <- (cpt +1) mod 6;
          valeur <- valeur +(cpt/5);
          incx <- -valeur;
          incy <- if incy>0 then valeur else -valeur;
          pointp2 <- pointp2 +1;
          beep();
          if pointp2 = 21 then pointp2 <- 0;
              
        end;
      pposx <- posx;
      posx <- posx + incx;
      
      posy <- posy + incy;
      
      if valeur > 10 then valeur <- 4;
      
      Thread.delay(0.02);
      {|putpos(posx,posy,pointp1,pointp2)}
    done;
end;;

class joueur(d,n,name,name_adv) as self = 
  val mutable position = dep_y
  val mutable old_pos = dep_y
  val mutable old_pos_adv = dep_y
  val mutable old_posx_bal = mil_x
  val mutable old_posy_bal = mil_y
  val mutable points = -1
  val mutable adv_points = -1
  val numero = n
  val display = d
  val nom = "Player:"^name
  val mutable nom_x = 0;
  val adv_nom = "Player:"^name_adv
  val mutable adv_nom_x = 0;
  val posx = if n=0 then min_x else max_x
  val couleur = if n=0 then blue else red
  val couleur_adv = if n=0 then red else blue
  
  method partir(l) = go (l) 

  method beep() = sound 440 10
  
  method nvellepos (putpos) = 
    let def t!() = {putpos(posx,position)} in {|t()} 
  
  method etat(p) = 
    set_color black;
    fill_rect nom_x (etat_y) 60 50;    
    moveto nom_x (etat_y);
    set_color couleur;
    draw_string (string_of_int(p));
  
  method etat_adv(p) =
    set_color black;
    fill_rect adv_nom_x (etat_y) 60 50;
    moveto adv_nom_x  (etat_y);
    set_color couleur_adv;
    draw_string (string_of_int(p));    
  
  method affiche (getpos,get_balle) =
    let c = foreground in
    let (posx_adv,position_adv) = getpos()  in
    let (posx_bal,posy_bal,pp1,pp2) = get_balle() in
    
    let pp1,pp2 =
      if pp1 = 21 || pp2 = 21 then
        0,0
      else
        pp1,pp2
    in
    let mespts = if numero=0 then pp1 else pp2
    and sespts = if numero=0 then pp2 else pp1 in
    
   (* redessine partiellement l'ecran *)
    set_color white;
    set_line_width 2;
    moveto 0 (terrain_H+offset_H);
    lineto ecran_L (terrain_H+offset_H);
    moveto 0 (offset_H);
    lineto ecran_L (offset_H);
    
    
    
    if (points <> mespts) then
      begin
        points <- mespts;
        self#etat(points);
      end;
    
    if (adv_points <> sespts) then
      begin
        adv_points <- sespts;
        self#etat_adv(adv_points);
      end;
    
    (* la balle *)
    if (old_posx_bal<>posx_bal)&&(old_posy_bal<>posy_bal) then
      begin
        set_color black;
        fill_circle old_posx_bal (old_posy_bal+offset_H) 5;
        set_color yellow;
        fill_circle posx_bal (posy_bal+offset_H) 5;
        old_posx_bal <- posx_bal;
        old_posy_bal <- posy_bal;
        set_color c;
      end;

    (* la barre adverse *)
    if old_pos_adv<> position_adv then
      begin
        set_color black;
        fill_rect posx_adv (old_pos_adv+offset_H) barre_E barre_L;
        set_color couleur_adv;
        fill_rect posx_adv (position_adv+offset_H) barre_E barre_L;
        old_pos_adv <- position_adv    
      end;

    (* la barre du joueur *)
    if old_pos <> position then
      begin
        set_color black;
        fill_rect posx (old_pos+offset_H) barre_E barre_L;
        set_color couleur;    
        fill_rect posx (position+offset_H) barre_E barre_L;
        old_pos <- position;
      end;
  
  
  
  method init(getpos,get_balle) = 
    open_graph (display^" "^string_of_int(ecran_L)^"x"^string_of_int(ecran_H)); 
    set_font "lucidasans-bold-12";
    set_color black;
    fill_rect 0 0 ecran_L ecran_H;
    set_color couleur;
    moveto 10 etat_y;
    draw_string nom;
    nom_x <- fst (text_size nom) + 15;
    set_color couleur_adv;
    moveto adv_x etat_y;
    draw_string adv_nom;
    adv_nom_x <- fst (text_size adv_nom) + adv_x + 15;
    self#affiche(getpos,get_balle)
  
  method jouer (putpos,getpos,get_balle) =
    let new_pos = ref (mouse_pos()) in  
    position <- !new_pos ;  
    while true do
      if ( (!new_pos) <> position ) then
        begin
          position <- !new_pos;
          self#nvellepos(putpos)
        end;
      self#affiche(getpos,get_balle);
      Thread.delay 0.02;
      new_pos := mouse_pos();
    done;
end;;





(*------------------------------------------------------------------------------*)


let def pong!(l,name,d) | pas_de_joueur!() =
    {attente_partenaire(l,name,d)}
  or pong!(l1,nom1,d1) | attente_partenaire!(l2,nom2,d2) = 
    let p1 = new joueur(d1,0,nom1,nom2)
    and p2 = new joueur(d2,1,nom2,nom1) 
    and boule = new balle(mil_x,mil_y) in
    
    
    let loc server def get_pos_p1() | pos_p1!(x,y) = pos_p1(x,y)| reply x,y
        or get_pos_p2() | pos_p2!(x,y) = pos_p2(x,y) | reply x,y 
        or put_pos_p1!(new_x,new_y) | pos_p1!(x,y) = pos_p1(new_x,new_y)
        or put_pos_p2!(new_x,new_y) | pos_p2!(x,y) = pos_p2(new_x,new_y) 
        or put_balle!(new_x,new_y,pp1,pp2) | pos_balle!(x,y,ptp1,ptp2) = pos_balle(new_x,new_y,pp1,pp2) 
        or get_balle() | pos_balle!(x,y,pp1,pp2) = pos_balle(x,y,pp1,pp2) | reply x,y,pp1,pp2 
          or beep() = {
                        {(p1#beep());}
                        |
                        {(p2#beep());}
                        |reply}       do { go l1; } in
   
    {
      (p1#partir(l1));
      (p1#init(get_pos_p2,get_balle) );
      (p1#jouer(put_pos_p1,get_pos_p2,get_balle) );}      
  |
    {
      (p2#partir(l2));
      (p2#init(get_pos_p1,get_balle));
      (p2#jouer(put_pos_p2,get_pos_p1,get_balle));}
  | {(boule#partir(l1);boule#jouer(put_balle,get_pos_p1,get_pos_p2,beep));}
  | pas_de_joueur() 
  | pos_p1(min_x,dep_y)
  | pos_p2(max_x,dep_y)
  | pos_balle(mil_x,mil_y,0,0);;  

{| pas_de_joueur()};;

let def new_game (l,name,_) = { pong (l,name,"") | reply}

let _ = 
  Arg.parse (Genserver.ns_args ()) (fun _ -> ()) "start Pong server";
  Genserver.server (new_game: (Join.location * string * string list -> unit)) vartype
  
