/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: libgraph.h,v 1.2 1999/11/24 13:22:32 lefessan Exp $ */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <mlvalues.h>

struct canvas {
  int w, h;                     /* Dimensions of the drawable */
  Drawable win;                 /* The drawable itself */
  GC gc;                        /* The associated graphics context */
};

struct rect_cellule{
  int w, h;
  int x, y;
  struct rect_list *suite;
};
typedef struct rect_cellule *rect_list;

extern Display * grdisplay;     /* The display connection */
extern int grscreen;                   /* The screen number */
extern Colormap grcolormap;            /* The color map */
extern struct canvas grwindow;         /* The graphics window */
extern struct canvas grbstore;         /* The pixmap used for backing store */
extern int grwhite, grblack;           /* Black and white pixels */
extern int grx, gry;                   /* Coordinates of the current point */
extern unsigned long grcolor;          /* Current drawing color */
extern XFontStruct * grfont;    /* Current font */

#define Wcvt(y) (grwindow.h - 1 - (y))
#define Bcvt(y) (grbstore.h - 1 - (y))
#define WtoB(y) ((y) + grbstore.h - grwindow.h)
#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define DEFAULT_SCREEN_WIDTH 600
#define DEFAULT_SCREEN_HEIGHT 450
#define BORDER_WIDTH 2
#define WINDOW_NAME "Caml Light graphics"
#define ICON_NAME "Caml Light graphics"
#define DEFAULT_EVENT_MASK \
          (ExposureMask | KeyPressMask | StructureNotifyMask)
#define DEFAULT_FONT "fixed"
#define SIZE_QUEUE 256

/* To handle events asynchronously */
#ifdef HAS_ASYNC_IO
#define USE_ASYNC_IO
#define EVENT_SIGNAL SIGIO
#else
#ifdef HAS_SETITIMER
#define USE_INTERVAL_TIMER
#define EVENT_SIGNAL SIGALRM
#else
#define USE_ALARM
#define EVENT_SIGNAL SIGALRM
#endif
#endif

extern void gr_fail();
extern void gr_check_open();
extern unsigned long gr_pixel_rgb();
extern int gr_rgb_pixel();
extern void gr_handle_simple_event();
extern void gr_enqueue_char();
extern void gr_init_color_cache();
extern int gr_update;
extern rect_list gr_listupdate;
extern void rl_inserer(int,int,int,int);
extern void rl_liberer();
