/***********************************************************************/
/*                                                                     */
/*                           Join-Calculus                             */
/*                                                                     */
/*            Fabrice Le Fessant, projet PARA, INRIA Rocquencourt      */
/*                                                                     */
/*  Copyright 1998 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "weak.h"
#include "gc.h"

#include "join.h"

/*********

   module Channel

*********/


void print_spaces(int offset)
{
  int i;
  for(i=0; i<offset; i++) printf(" ");
}


void print_iter(value v, int offset, int depth)
{
  print_spaces(offset);
  
  if(depth < 1){
    printf("...\n"); return;
  }
  
  if(Is_block(v)){
    if(Is_in_heap(v) || Is_young(v)){
      int tag = Tag_val(v);
      int size = Wosize_val(v);
      int i; 

      if(tag >= No_scan_tag){
        printf("[(%d) %d: ___] \n", tag, size); return;
      }
      
      switch(tag){
        case String_tag:
        printf("str = %s\n", String_val(v)); break;

        case Double_array_tag:
        case Double_tag:
        case Final_tag:
        printf("[(%d) %d: ___] \n", tag, size); break;

        default:      
        printf("[(%d) %d:\n", tag, size);
        for(i=0; i<size; i++){
          print_spaces(offset);
          printf("Field %d\n", i);          
          print_iter(Field(v,i), offset+2, depth-1);
        }
        print_spaces(offset);
        printf("]\n");
      }
    } else 
      if (Is_atom(v)){
        printf("Atom %d\n",Tag_val(v));
      } else {
        printf("??? = %lx\n", v);
      }
  } else {
    printf("int = %d\n", Int_val(v));
  }
}

value print_obj(value v)
{
  print_iter(v, 0, 5);
  return Val_unit;
}


extern value native_closure;

value sync_channel = Val_unit;
value async_channel = Val_unit;
int sync_channel_size;
int async_channel_size;

value register_send_sync(value sync_closure,value async_closure)
{
  sync_channel = sync_closure;
  register_global_root(&sync_channel);
  async_channel = async_closure;
  register_global_root(&async_channel);
  sync_channel_size = Wosize_val(sync_channel);
  async_channel_size = Wosize_val(async_channel);
/*  fprintf(stderr,"Closures: %d - %d\n", sync_channel_size,async_channel_size); */
  
  return Val_unit;
}

value unbox_channel(value channel)
{
  value res;

  Begin_root(channel);
  res = alloc(2,0);
  Field(res,0) = Field(channel,sync_channel_size-2);
  Field(res,1) = Field(channel,sync_channel_size-1);
  End_roots();

  return res;
}
 
value extract_stub(value obj)
{
  if(Is_block(obj)){
    switch(Tag_val(obj)){
    case Object_tag: return obj;
    case Location_tag: return obj;
    case Thread_tag: return obj;
    case Automata_tag: return obj;
    case Closure_byte_tag:
      
      if(Wosize_val(obj) == async_channel_size &&
         Is_block(Field(obj,async_channel_size-2)) &&
         Tag_val(Field(obj,async_channel_size-2)) == Automata_tag)
        return Field(obj,async_channel_size-2);
    default:;
    }
  }
  
  failwith ("fail of something not a reference");
  return Val_unit;
}

  
/* On ajoute un mask dans la structure de match: 
     state = new_state XOR mask.
   Les <- sont transformes en un trigger_def sans empilement.

let def s(n,m) | a! x = s <- (n+x,m) | reply

est transforme en:

let def _s(_) | a! x = let (n,m) = get_state s in set_state s (n+x,m) | reply

   */
   
/* return: bloc[automata,channels...] */


#define CLASS_TABLE 2
#define TABLE_EXT_TMV 2

value create_empty_automata(value location, value obj_class, value nchans_v,
			       value indexes,value syncs) /* ML */
{
  value tmp,i,j,ldef;
  value nchans = Int_val(nchans_v);
  value res=Val_unit;
  value def=Val_unit;
  int stub_size;

  if(obj_class == Val_unit) stub_size = STUB_SIZE;
  else stub_size = STUB_SIZE + 1;

  if(location == Val_unit)
    location = (value)curr_thread->location;

  Begin_roots4(location,obj_class, indexes,syncs);  /* (res,def,ldef) */
  res = obj_block(Val_int(0),Val_int(Wosize_val(indexes)+1));
  Begin_root(res);
  def = obj_block(Val_int(Automata_tag),Val_int(stub_size));
  Begin_root(def);
  ldef = obj_block(Val_int(LocalAutomata_tag),Val_int(LDEF_OFFSET+2*nchans));

  Assign(Field(ldef,LDEF_LOCATION), location);
  Assign(Field(def,STUB_POSITION), Field(location,STUB_POSITION));
  Assign(Field(def,STUB_LOCAL),ldef);
  Assign(Field(res,0),def);

  if(obj_class != Val_unit){
    Assign(Field(def, STUB_EXT_TMV), 
      Field(Field(obj_class, CLASS_TABLE), TABLE_EXT_TMV));
  }

  for(i=0;i<Wosize_val(indexes);++i){
    if(Field(syncs,i) == Val_true){
      tmp = alloc(sync_channel_size,Closure_byte_tag); /* les canaux */
      for(j=0;j<sync_channel_size-2;j++)
	Field(tmp,j) = Field(sync_channel,j);
    } else {
      tmp = alloc(async_channel_size,Closure_byte_tag); /* les canaux */
      for(j=0;j<async_channel_size-2;j++)
	Field(tmp,j) = Field(async_channel,j);
    }
    Field(tmp,j++) = def;
    Field(tmp,j) = Field(indexes,i);
    Assign(Field(res,i+1),tmp);
  }

  tmp = weakset_add(((location_t)location)->lloc->def_weakset,def);
  Assign(((location_t)location)->lloc->def_weakset,tmp);

  End_roots();
  End_roots();
  End_roots();
    
  return res;
}

value fill_def(value def,value processes,value structure) /* ML */
{
  value ldef = Field(def,STUB_LOCAL);
  
  Assign(Field(ldef,LDEF_STRUCTURE),structure);
  Assign(Field(ldef,LDEF_PROCESSES),processes);
  
  return Val_unit;
}

value trigger_def (value params)
{
  value def = Field(params,0); /* BLOCK */
  value ord_v = Field(params,1);
  value arg = Field(params,2);  /* BLOCK */
  value ord = Int_val(ord_v);
  value index = ord & 127;
  value dispatch = ord >> 10;

  value ldef = Field(def,STUB_LOCAL); /* BLOCK */
  value structure = Field(ldef,LDEF_STRUCTURE);   /* BLOCK */
  value nmult = Int_val(Field(structure,STRUCT_NMULT));
  value state = Field(ldef,LDEF_STATE);
  value bit = 1 << (index + 1);
  value cons;

  Good_stub("trigger_def",def);

  if(! (state & bit)){    /* pas ce message */
    value tests = Field(structure,STRUCT_OFFSET+index); /* BLOCK */
    value ntests =   Wosize_val(tests);
    value new_state = state | bit;
    value inverse = ~new_state;
    value nchans = Int_val(Field(structure,STRUCT_NCHANS)); 
    value i;
    value processes = Field(ldef,LDEF_PROCESSES);   /* BLOCK */

    for(i=0;i<ntests;i+=2){
      if(!(Field(tests,i) & inverse)){
      /* we must trigger this process */
	/* first we have to construct its args */
	value match = Field(tests,i+1);
	value nargs;
	value process;
	value thread_num;

	/* un test de trop: ca peut etre fait definitivement au premier 
	   fill_automata */
	if(Is_long(match)){
	  match = Field(structure,STRUCT_OFFSET+nchans+Int_val(match));
	  Assign(Field(tests,i+1),match);
	}
	nargs = Wosize_val(match)-1;
	process = Field(match,0);
	
	if(nargs==1){ /* un seul argument, c'est le message ! */
	  if(index < nmult) /* dispatch */
	    process = Field(process,dispatch);

	  thread_num = Int_val(Field(process,1));
	  process = Int_val(Field(process,0));
	  if (!thread_num){
	    gen_thread_new(Field(ldef,LDEF_LOCATION),
			   Field(processes,process),
			   Val_true,arg);
	  } else 
            if(thread_num!=-1){
              reuse_thread_loc(Field(ldef,LDEF_LOCATION),
                               Field(processes,process),
                               arg,Field(arg,0));
            } else {
              /* argh ! this thread should  never be preempted, else
              a migration could happen and the two threads would be 
              in conflict  */
              local_thread_reuse(Field(ldef,LDEF_LOCATION),
                                 Field(processes,process),
                                 arg,curr_thread);
            }
          return Val_unit;
	} else {
	  value args;
	  value chan;
	  value list;

	  Begin_roots5(match,processes,process,arg,ldef);
	  args = alloc(nargs,0);
	  End_roots();
	  for(i=0;i<nargs;i++){
	    chan = Int_val(Field(match,i+1));
	    if(chan==index){ /* message actuel */
	      if(index<nmult) /* dispatch */
		process = Field(process,dispatch);
	      Assign(Field(args,i),arg);
	    } else { /* message queue */
	      list = Field(ldef,LDEF_OFFSET+chan);
	      if(chan<nmult) /* dispatch */
		process = Field(process, Int_val(Field(list,2)));
	      Assign(Field(args,i),Field(list,0));
	      list = Field(list,1);
	      if(list == Val_unit)
		state &= ~(1 << (chan+1));
	      Assign(Field(ldef,LDEF_OFFSET+chan),list);
	    }
	  }
	  Field(ldef,LDEF_STATE) = state;
	  thread_num = Int_val(Field(process,1));
	  process = Int_val(Field(process,0));
	  if (!thread_num){
	    gen_thread_new(Field(ldef,LDEF_LOCATION),
			   Field(processes,process),
			   Val_true,args);
	  } else
            if(thread_num != -1) {
              value thread_stub = Field(Field(args,thread_num-1),0);
              reuse_thread_loc(Field(ldef,LDEF_LOCATION),
                               Field(processes,process),
                               args,
                               thread_stub);
            } else {
              local_thread_reuse(Field(ldef,LDEF_LOCATION),
                                 Field(processes,process),
                                 args,curr_thread);
            }
	  return Val_unit;
	}
      }
    }
    Field(ldef,LDEF_STATE) = new_state;
  }
  Begin_roots2(ldef,arg); /* ldef, arg */
  if(index < nmult){
    cons = alloc(3,0);
    Field(cons,2) = Val_int(dispatch);
  } else
    cons = alloc(2,0);
  Field(cons,0) = arg;
  Field(cons,1) = Field(ldef,LDEF_OFFSET+index);
  Assign(Field(ldef,LDEF_OFFSET+index),cons);
  End_roots();

  return Val_unit;
}

value marshal_def (value def) /* ML */
{
  value tmp = Field(def,STUB_LOCAL);
  
  Assign(Field(def,STUB_LOCAL),Val_unit);
  Field(def,STUB_CACHE) = Val_unit;

  return tmp; /* ldef */
}

value unmarshal_def (value paire) /* ML */
{ /* paire = (def,ldef) */
  value lloc;
  value tmp;
  value def = Field(paire,0);
  value ldef = Field(paire,1);
  
  Assign(Field(def,STUB_LOCAL), ldef);
  Field(def,STUB_CACHE) = Val_unit;
  
  lloc = Field(Field(ldef,LDEF_LOCATION),STUB_LOCAL);
  Begin_root(lloc);
  tmp = weakset_add(Field(lloc,LLOC_DEFWEAKSET),def);
  Assign(Field(lloc,LLOC_DEFWEAKSET),tmp);
  End_roots();
  
  return Val_unit;
}

#ifdef DEBUG
#define Good_channel(msg,channel) good_channel(channel,msg,__FILE__,__LINE__)
#else
#define Good_channel(msg,channel)
#endif

void good_channel(value channel,char* msg,char *file,int line)
{
  if(!Is_block(channel)) failed_message(msg, "Channel is not a block",file,line);
  fprintf(stderr,"Channel value %s %x", msg, channel);
  if(Tag_val(channel)!=Automata_tag)
    failed_message(msg,"Not an automata",file,line);
    
/*   good_stub(channel,msg,file,line); */
}


value test_good_channel(value channel,value string)
{
  
  Good_channel(String_val(string),channel);
  
  return Val_unit;
}

value print_channel (value str, value v)
{
  fprintf(stderr, "Print %s : %x",String_val(str), v);
  return Val_unit;
}

void check (value ptr, int deep)
{
   int size, tag,i;

  /* Here, we can check the minor heap (young_ptr <-> young_end) */


  if(!(Is_block(ptr) && (Is_in_heap(ptr) || Is_young(ptr)))) return;
  size = Wosize_val(ptr);
  tag = Tag_val(ptr);

  if(size > 10000) { fprintf(stderr, "Suspected size %d\n",size); exit(2); }
  if (tag < No_scan_tag || deep > 0) return;
  for(i=0; i< size; i++){
    check(Field(ptr, i),deep+1);
  }
}

value check_pointer (value ptr, value i)
{
  check(ptr,0);

  return Val_unit;
}
