/***********************************************************************/
/*                                                                     */
/*                           Join-Calculus                             */
/*                                                                     */
/*            Fabrice Le Fessant, projet PARA, INRIA Rocquencourt      */
/*                                                                     */
/*  Copyright 1998 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "weak.h"
#include "gc.h"

#include "join.h"

/*********

   module Location

*********/
/*
external create_new_object : space -> location -> int -> Reloc.module_data ->
  Obj.t = "create_new_object";;
*/

value set_location_module(value obj, value module)
{
  Assign(Field(Field(obj,STUB_LOCAL),LLOC_MODULE),module);
  return Val_unit;
}

value get_location_module(value obj)
{
  return Field(Field(obj,STUB_LOCAL),LLOC_MODULE);
}

value create_location(value space,value father,value module_data,
		      int size, int tag)
{
  value loc,lloc;
  value tmp;

  Begin_roots3(space,father,module_data);
  /* (space,father,module_data,ll) */

  lloc = obj_block(Val_int(LocalLocation_tag),Val_int(LLOC_SIZE));

  Begin_root(lloc);

  Assign(Field(lloc,LLOC_FATHER),father);
  tmp = weakset_create(Val_int(DEF_WEAKSET_SIZE));
  Assign(Field(lloc,LLOC_DEFWEAKSET),tmp);
  tmp = weakset_create(Val_int(LOC_WEAKSET_SIZE));
  Assign(Field(lloc,LLOC_LOCWEAKSET),tmp);
  Assign(Field(lloc,LLOC_MODULE),module_data);
  
  loc = obj_block(Val_int(tag),Val_int(size));
  
  Assign(Field(loc,STUB_LOCAL),lloc);
  Assign(Field(loc,STUB_POSITION),space);

  if(father != Val_unit){
    Begin_root(loc); /* (l) */
    tmp = weakset_add(Field(Field(father,STUB_LOCAL),LLOC_LOCWEAKSET),loc);
    Assign(Field(Field(father,STUB_LOCAL),LLOC_LOCWEAKSET),tmp);
    End_roots(); /* (l) */
  }
  End_roots(); /* (space,father,module_data,ll) */
  End_roots();
  
  return loc;
}

value create_new_location(value space,value father, 
			  value module_data) /* ML */
{
  return create_location(space,father,module_data,STUB_SIZE, Location_tag);
}

/*
 
     STUB
   5 TMV externe
   6 TMV interne
   variables d'instances

*/

value remove_from_father(value location)
{
  value lloc = Field(location,STUB_LOCAL);

  weakset_remove(Field(Field(Field(lloc,LLOC_FATHER),STUB_LOCAL),
		       LLOC_LOCWEAKSET),location);

  return Val_unit;
}

#define IN_BUCKET 6
value marshal_location(value location) /* ML */
{
  value res;
  value tmp;
  value size = Wosize_val(location);
  value lloc = Field(location,STUB_LOCAL);
    
  Begin_roots2(lloc,location);
  res = obj_block(Val_int(0),Val_int(6));
  Begin_root(res);

  Field(location,STUB_CACHE) = Val_unit;

/* Il faut faire un test pour savoir si la location n'est pas failed avant
d' essayer de la marshaller ... */

  weakset_remove(Field(Field(Field(lloc,LLOC_FATHER),STUB_LOCAL),
		       LLOC_LOCWEAKSET),location);
  
  Assign(Field(res,3),Field(lloc,LLOC_MODULE));
  Assign(Field(res,4),Field(lloc,LLOC_ATTACH));
  Field(res,5) = Atom(0);

  tmp = extract_location_threads(location);
  Assign(Field(res,0),tmp);  /* les threads */
  tmp = weakset_array(Field(lloc,LLOC_DEFWEAKSET));
  Assign(Field(res,1),tmp);  /* les defs */
  tmp = weakset_array(Field(lloc,LLOC_LOCWEAKSET));
  Assign(Field(res,2),tmp);  /* les sous-locations */

  /* si la location est un objet */
  if(size > STUB_SIZE){
    value vars = alloc(size-IN_BUCKET,0); 
    int i;

    for(i=IN_BUCKET;i<size;i++){
      Init(Field(vars,i-IN_BUCKET),Field(location,i));
    }
    /* IN_BUCKET+1 since the local TMV is used in Jmarshal */
    for(i=IN_BUCKET+1;i<size;i++){ Assign(Field(location,i),Val_unit);  }
    Init(Field(res,5),vars);
  }

  End_roots();
  End_roots();
  
  return res;
}

value unmarshal_location (value father,value location,
			  value module_data,value attach, 
			  value vars)   /* ML */
{
  value lloc;
  value tmp;
 
  Field(location,STUB_CACHE) = Val_unit;

  Begin_roots5(father,location,module_data,attach,vars);

  if(Wosize_val(location) > STUB_SIZE){
    int size = Wosize_val(vars);
    int i;
    /* il s'agit d'un objet */
    for(i=0;i<size;i++)
      Assign(Field(location,i+IN_BUCKET),Field(vars,i));
  }
  
  tmp = weakset_add(Field(Field(father,STUB_LOCAL),LLOC_LOCWEAKSET),location);
  Assign(Field(Field(father,STUB_LOCAL),LLOC_LOCWEAKSET),tmp);
  Assign(Field(location,STUB_POSITION),Field(father,STUB_POSITION));
    

  lloc = obj_block(Val_int(LocalLocation_tag),Val_int(LLOC_SIZE));
  Assign(Field(location,STUB_LOCAL),lloc);
  Assign(Field(lloc,LLOC_FATHER),father);
  Assign(Field(lloc,LLOC_MODULE),module_data);
  Assign(Field(lloc,LLOC_ATTACH),attach);
  
  Begin_root(lloc);
  tmp = weakset_create(Val_int(DEF_WEAKSET_SIZE));
  Assign(Field(lloc,LLOC_DEFWEAKSET),tmp);
  tmp = weakset_create(Val_int(LOC_WEAKSET_SIZE));
  Assign(Field(lloc,LLOC_LOCWEAKSET),tmp);
  
  End_roots();
  End_roots();

 return Val_unit;
}

/*
  On attache une unite de relocation a une location, sous la forme
d'une liste de (unite de relocation,interface). On migre la location avec 
ses attaches, puis, quand on la demarshalle, on place les interfaces sur
les unites de relocation, permettant ainsi les appels non resolus.
 */
value attach_to_location(value module_data) /* ML */
{
  value cons;
  value lloc;
  
  Begin_root(module_data);
  cons = alloc(2,0);
  lloc = Field(curr_thread->location,STUB_LOCAL);
  Field(cons,0) = module_data;
  Field(cons,1) = Field(lloc,LLOC_ATTACH);
  Assign(Field(lloc,LLOC_ATTACH),cons);
  
  End_roots();
  return Val_unit;
}

value schedule_critical_thread(int action,value arg1,value arg2)
{
  value res;
  
  Begin_roots2(arg1,arg2);
  res = alloc(2,action);
  
  Field(res,0) = arg1;
  Field(res,1) = arg2;

  End_roots();
  return schedule_go_thread(res);
}

value migrate(value orig,value dest) /* ML */
{
  location_t l;
  location_t self = (location_t) orig;
  value tmp;
  
  l=(location_t)dest;
  
  if(self->lloc->father == l) return Val_unit; /* deja fait */
  if((value)(self->lloc->father) == Val_unit)
    failwith("Go: migration of space location");

  if(Field(dest,STUB_LOCAL) != RemoteObject_val &&
     Tag_val(Field(dest,STUB_LOCAL)) == LocalLocation_tag
     ){
    /* migration locale */
    while((value)l!=Val_unit){
      if(l == self) 
	failwith("Go: migration to children location");
      l=l->lloc->father;
    }
    l=(location_t)dest;
    if(weakset_remove(self->lloc->father->lloc->loc_weakset,(value)self)
       ==Val_false)
      failwith("Local Go: Unable to remove location");
    Assign(self->lloc->father,l);
    Begin_root(l);
    tmp = weakset_add(l->lloc->loc_weakset,(value)self);
    Assign(l->lloc->loc_weakset,tmp);
    End_roots();
    return Val_unit;
    
  } else

    return schedule_critical_thread(0,orig,dest);
}

value kill_location(value location)
{
  return schedule_critical_thread(1,location,Val_unit);
}

value fail_space(value space) /* Normal -> MaybeDown */
{
  return schedule_critical_thread(2,space,Val_int(1));
}

value halt_space(value space) /* Normal -> Down */
{
  return schedule_critical_thread(2,space,Val_int(2));
}

value go_location(value dest)
{
  return migrate(curr_thread->location,dest);
}

value migrate_location(value pair)
{ 
  return  migrate(Field(pair,0),Field(pair,1));
}

/*********

   module Interfaces

 ********/
value get_interface(value module_data) /* ML */
{
  if(Field(module_data,INTERFACE) == Val_int(-1))
    failwith("Location attached to unset Module");
  return Field(global_data,Int_val(Field(module_data,INTERFACE)));
}

value next_free_global = 0;
value end_free_global = 0;

extern void realloc_global(value);

value set_interface(value module_data,value interface) /* ML */
{
  if(Field(module_data,LONGUEUR) == Val_int(0))
    failwith("Can't attach interface to dummy module");
  if(Field(module_data,INTERFACE) != Val_int(-1)) 
    return Val_unit; /* already done */

  /* il faut ajouter un slot a la table globale pour l'interface.
     Ceci n'est possible que si on ne se trouve pas dans le toplevel (qui
     manipule aussi la table globale ! ).
   */

  if(next_free_global == end_free_global){
    /* no slots reserved for interfaces */
    next_free_global = Wosize_val(global_data);
    end_free_global = next_free_global + 20;
    realloc_global(Val_int(end_free_global));
  }
  Assign(Field(global_data,next_free_global),interface);
  Field(module_data,INTERFACE) = Val_int(next_free_global++);
  
  return Val_unit;
}
