open Mop
  
(* Object copy:
     En fait, il faudrait faire plus: lorsque l'on demande a obtenir un 
replicat, on s'attend a ce que les threads que l'on a lances sur l'original
soient donnes au replicat. Ainsi, en cas de rupture de liaison avec l'original,
les invocations lancees sur l'original sont aussi dans le cache. Un probleme 
de coherence se pose cependant.
   La question est: Si je possede un replicat d'un objet, mes invocations 
doivent-elles aller quand-meme sur l'original pour un maximum de coherence,
ou dois-je utiliser mon replicat, dans le cache, qui lui maintient une
coherence reguliere. Si je lance systematiquement mes threads sur le replicat,
la coherence dans le travail semble plus forte.
 *)

let copy (o: < .. >) = 
  let o' = Obj.repr o in
  let (locator, position, local, _, gc_stub) = Obj.magic o' in
  if (Obj.magic local) == None then
    failwith "Can not copy non local object";
  let sz = Obj.size o' in
  let r = 
    let self_loc = Jprims.self_location () in
    Jprims.create_new_object Space.local_space 
      self_loc sz (Jprims.get_location_module o)
  in
  for i = 5 to sz - 1 do Obj.set_field r i (Obj.field o' i) done;
  Obj.obj r 

let new_method = Mop.new_method

let set_initializer = Mop.set_initializer
let inheritance = Mop.inheritance
let get_method = Mop.get_method
let set_method = Mop.set_method
let get_method_label = Mop.get_method_label
let get_variable = Mop.get_variable
let hide_variable = Mop.hide_variable
let get_private_variable = Mop.get_private_variable
let create_class = Mop.create_class
let get_class_info = Mop.get_class_info

(* Parameters *)
let params = Mop.params

let local_classes = Hashtbl.create 13
  
(* Objects *)
let create_object table =
  let class_info = (Obj.magic table.buckets.(0) : class_info) in
  let class_id = class_info.class_type in
  let (Class_id(module_name,class_no)) = class_id in
  let class_info = try
      Hashtbl.find local_classes class_id
    with _ -> 
        Hashtbl.add local_classes class_id class_info;
        class_info.table.ext_buckets 
        <- Channel.create_ext_buckets class_info.table.messages ;
        class_info
  in
  let table = class_info.table in
  let module_data = Reloc.get_module_data module_name in
  let obj = 
    let self_loc = Jprims.self_location () in
    Jprims.create_new_object Space.local_space 
      self_loc table.size module_data
  in
  Obj.set_field obj 5 (Obj.repr table.ext_buckets);     (* TMV *)
  Obj.set_field obj 6 (Obj.repr table.buckets);     (* TMV *)
  let initialization = Obj.magic List.hd (List.hd table.init) in
  (* sur un autre thread *)
  Jprims.set_critical_section  true;
  let thread_stub = Jprims.self_stub () in
  Jprims.thread_create_in (Obj.magic obj:Jtypes.location_stub) 
  (function () -> (* sans interet, on retourne une closure qui ne s'executera donc
  pas dans l'objet *)
        Channel.reply_to_and_exit thread_stub
          (try
            Jtypes.Reply 
            (Obj.magic (Mop.constructor (initialization obj) (Obj.magic obj) table))
        with
          e -> Jtypes.Exception e)
        ) ();
  Jprims.set_critical_section false;
  match  Jprims.wait_on_stub () with
  | Jtypes.Reply v -> Obj.magic v
  | Jtypes.Exception e -> raise e

      
let create_class class_info public_methods creator class_init module_data =
  Debug.print0 'o' "Mop.create_class...";
  let class_info =
    Mop.create_class class_info public_methods creator class_init module_data
  in
(* ajouts des methodes predefinies dans la TMV *)
  class_info.class_sig <- class_info.table.messages;
  Debug.print0 'o' "...Mop.create_class\n";
  (* il faut reconstruire class_info.table.ext_buckets a partir de
       class_info.table.buckets et de class_info.table.messages
  *)
  Debug.print0 'o' "...Oo.create_class\n"

