open Jtypes
open Metatype

exception TypeMismatch of mtype_expr

let rec print_list print n sep elist = 
    match elist with
      [] -> ""
    | hd::tl ->
        Printf.sprintf "%s%s%s"
          (if n then sep else "")
        (print hd)
        (print_list print true sep tl)

  
let to_string mtype =
  let aliases = alias mtype
  in
  let rec iter expr =
    try
      let (v,p) = List.assq !expr aliases
      in
      if !p then Printf.sprintf "'%d" v else begin
          let s = Printf.sprintf "(%s) as '%d" (print expr) v in
          p := true;
          s
        end
    with
      Not_found ->
        print expr
        
  and print expr =
    match !expr with
      MTnil -> ""
    | MTarrow (e1,e2) -> Printf.sprintf "(%s -> %s)" (print e1) (print e2)
    | MTchannel e -> Printf.sprintf  "<<%s>>" (print e)
    | MTtuple elist -> Printf.sprintf "(%s)"
          (print_list print false " * " elist)
    | MTconstr (name,elist) -> 
        Printf.sprintf "%s %s"
          (match elist with
            [] -> ""
          | [hd] -> print hd
          | elist -> Printf.sprintf "(%s)" (print_list print false "," elist)) 
        (match name with
            MType_int -> "int"
          | MType_char -> "char"
          | MType_string -> "string"
          | MType_float -> "float"
          | MType_bool -> "bool"
          | MType_exn -> "exn"
          | MType_unit -> "unit"
          | MType_array -> "array"
          | MType_list -> "list"
          | MType_metatype -> "metatype"
          | MType_name (name,kind) ->  name)
    | MTobject list ->
        Printf.sprintf "<%s>" (print_list 
            (fun (s,e) -> Printf.sprintf "%s:%s" s (iter e)) false " ;" list)
  in
  iter mtype
  
let _ =
  Printexc.add_printer (fun e ->
      match e with
        TypeMismatch mt -> 
          Printf.sprintf "TypeMismatch: with type %s" (to_string mt)
      | _ -> raise Not_found
  )
  
let setLocalService name (object:'a) (vt:'a metatype) =
  let vt = Obj.repr vt in
  Jprims.critical_section (function () ->
    (try
      Hashtbl.find Space.name_server name;
      Hashtbl.remove Space.name_server name;
    with
      Not_found -> ());
    Hashtbl.add Space.name_server name ((Obj.repr object),vt);
    )

let getLocalService name (x:'a metatype) =
  let (o,y) =
    Obj.magic (
    Jprims.critical_section   (function () ->
      Hashtbl.find Space.name_server name))
  in
  let yy = translate y in
  if le yy (translate x)  then
    (o:'a)
  else begin
      Printf.printf "RAISING TYPEMISMATCH"; 
      Metatype.print_type yy;
      print_newline ();
      raise (TypeMismatch yy) 
    end

let getRemoteService space name (vt:'a metatype) =
  Jprims.critical_section (fun () ->
    Debug.print "get 2";
    Space.send space
      (let t =
         Jmarshal.marshal space 
          (GetService (name,(Obj.repr vt), Jprims.self_stub ()))
      in
      t);
    Debug.print "get 3";
    match Jprims.sleep () with (* Jprims.wait_on_stub *)
    | Reply v -> (Obj.magic v:'a)
    | Exception e -> raise e
          )
