(***********************************************************************)
(*                                                                     *)
(*                      Join Objective Caml                            *)
(*                                                                     *)
(*        Fabrice LE FESSANT, Project PARA, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typejoin.ml,v 1.10 2001/01/05 10:56:54 lefessan Exp $ *)

(* 
Translation of automata as object:

let class#1 = Join.create_automata_class [noms_des_methodes]
  
    ...
  
let (object, ...) =
  let automata = Join.create_automata ... in
  ...
  let object = Join.create_automata_object class#1 [noms_des_canaux] in
   (* L'objet est un objet, mais pas une location *)
*)



(* Typechecking for the core language *)

open Misc;;
open Asttypes;;
open Longident;;
open Parsetree;;
open Types;;
open Typedtree;;
open Btype;;
open Ctype;;
open Typemisc;;
open Typecore;;

type automata_object = {
    obj_in_ident : Ident.t;
    obj_ext_ident : Ident.t;
    obj_class : Ident.t;
  }

type automata = { 
    obj_name : automata_object option;
    obj_init : Parsetree.expression option;
    automata : Ident.t;
    automata_type : type_expr;
    location : Ident.t option; 
    channels : (string * channel) list;
    processes : (join * Parsetree.expression) list;
    syntax_loc : Location.t;
    nmult: int;
    nchans: int;
    f: int array;
    instvars: (string * channel) list;
  }

let error = jerror
let add_value name ty env =
  Env.add_value name { val_type = ty; val_kind = Val_reg } env

(* Used to detect safe tail synchrone: a synchronous channel, inside a
location *) 

let local_location = Ident.create "local_location" let
  current_location = ref local_location
  

let rec fields_of_channels channels ptype sign =
  match channels with
    [] -> ptype
  | (name, { chan_sync = chan_sync; chan_type = chan_type }) :: channels ->
      match sign with
        None ->
          fields_of_channels channels 
            (newty (Tfield ( name, Fpresent, 
                chan_type, ptype))) sign
      | Some list ->
          fields_of_channels channels 
            (if List.mem name list then
              (newty (Tfield ( name, Fpresent, 
                    chan_type, ptype)))
            else ptype)
          sign
          

let automata_classes = ref []
  
let rec compile_automata env 
    (location, (obj,def_list) ,syntax_loc) =
  let nchannels = ref (-1) in
  let ninstvars = ref 0 in
  let (channels, instvars, processes) =
    List.fold_left
      (fun (channels, instvars, processes) (pattern, process) ->
        let join =
          {join_pattern = []; join_args = []; join_vars = [];
            join_replies = []; join_id = Ident.create "#process#";
            join_types = []}
        in
        let rec iter (channels,instvars) pattern =
          match pattern.ppat_desc with
            Ppat_or (pat1, pat2) -> iter (iter (channels,instvars) pat2) pat1
          | Ppat_construct (Lident name, Some args, _ ) ->
(* OK une variable d'instance *)
              let _ = pattern_variables := [] in
              let (instvar, instvars) =
                try
                  (List.assoc name instvars,instvars)
                with
                  Not_found ->
                    decr ninstvars;
                    let tyl = newvar () in
                    let instvar =
                      { chan_ext_ident = Ident.create name;
                        chan_in_ident = Ident.create name;
                        chan_args = tyl;
                        chan_type = newty (Tchannel (tyl));
                        chan_types = []; 
                        chan_ord = !ninstvars; 
                        chan_index = 0;
                        chan_sync = None}
                    in
                    instvar, (name, instvar) :: instvars
              in
              if List.mem instvar.chan_ord join.join_pattern then
                error pattern.ppat_loc 
                  ("Twice channel "^name^" in one join pattern");
              join.join_pattern <- instvar.chan_ord :: join.join_pattern;
              let pat_args = type_pat env args in
              unify env pat_args.pat_type instvar.chan_args;
              join.join_args <- (instvar, pat_args) :: join.join_args;
              join.join_vars <- !pattern_variables @ join.join_vars;
              join.join_types <-
                List.map (fun (v,(ty,_)) -> ty) !pattern_variables 
                :: join.join_types;
              (channels,instvars)
          | Ppat_channel (name, pat, sync) ->
              let _ = pattern_variables := [] in
              let (channel, channels) =
                try
                  let channel = List.assoc name channels in
                  match sync, channel.chan_sync with
                    Sync, None | Async, Some _ ->
                      error pattern.ppat_loc 
                        ("Different synchrony for channel "^name)
                  | _ -> channel, channels
                with
                  Not_found ->
                    incr nchannels;
                    let tyl = newvar () in
                    let channel =
                      match sync with
                        Sync ->
                          let treply = newvar () in
                          {chan_ext_ident = Ident.create name;
                            chan_in_ident = Ident.create name;
                            chan_args = tyl;
                            chan_type = newty (Tarrow (tyl, treply));
                            chan_types = []; 
                            chan_ord = !nchannels; 
                            chan_index = 0;
                            chan_sync = Some treply }
                      |  Async ->
                          {chan_ext_ident = Ident.create name;
                            chan_in_ident = Ident.create name;
                            chan_args = tyl;
                            chan_type = newty (Tchannel (tyl));
                            chan_types = []; 
                            chan_ord = !nchannels; 
                            chan_index = 0;
                            chan_sync = None }
                    in
                    channel, (name, channel) :: channels
              in
              if List.mem channel.chan_ord join.join_pattern then
                error pattern.ppat_loc 
                  ("Twice channel "^name^" in one join pattern");
              join.join_pattern <- channel.chan_ord :: join.join_pattern;
              let args = pat in
              let pat_args = type_pat env args in
              let _ =
                unify env pat_args.pat_type channel.chan_args 
              in
              let pat_args =
                match channel.chan_sync with
                  None -> pat_args
                | Some ty ->
                    let tyj = join_type "reply_event" env in
                    let id =
                      enter_variable !synthetic_loc ("#reply#" ^ name) 
                      (ty,Val_reg)
                    in
                    join.join_replies <- (name, id) :: join.join_replies;
                    mkpattuple [mkpat (Tpat_var id) tyj; pat_args]
              in
              let _ =
                join.join_args <-
                  (channel, pat_args) :: join.join_args;
                join.join_vars <- !pattern_variables @ join.join_vars;
                join.join_types <-
                  List.map (fun (v,(ty,_)) -> ty) !pattern_variables 
                  :: join.join_types
              in
              (channels,instvars)
          | _ -> failwith "Unknown pattern in join pattern"
        in
        let (channels,instvars) = iter (channels,instvars) pattern in
        join.join_pattern <- Sort.list (<) join.join_pattern;
        List.iter (fun (pjoin,_) ->
            if pjoin.join_pattern = join.join_pattern then
              error syntax_loc "Twice processes with same synchronization"
        ) processes;
        channels, instvars, (join, process) :: processes)
    ([], [], []) def_list
  in
  let channels = 
    Sort.list (fun (_,x) (_,y) -> x.chan_ord <= y.chan_ord) (channels) 
  in
  let nmult,nchans,f = Janalyse.fusion (List.length channels) 
    (List.map (fun (join,_) -> join.join_pattern) processes) in
  let automata, obj_name, obj_init, automata_type = match obj with
      None -> 
        let automata_type = mkmodty env "Join" "automata" in
        Ident.create "#automata#", None, None, automata_type
    | Some (obj_name, exp, sign) ->
        let obj_in_ident = Ident.create obj_name in
        let obj_ext_ident = Ident.create obj_name in
        let obj_type =
          newty (Tobject (
              fields_of_channels channels (newty Tnil) sign, ref None))
        in
        let obj_init = match exp.pexp_desc with
            Pexp_null -> None
          | _ -> Some exp
        in
        let obj_class = Ident.create "#automata_class#" in
        automata_classes := (obj_class, channels) :: !automata_classes;
        obj_in_ident, Some 
          { 
          obj_in_ident = obj_in_ident;
          obj_ext_ident  = obj_ext_ident;
          obj_class = obj_class;
        }, obj_init, obj_type

  in
  { 
    obj_name = obj_name;
    obj_init = obj_init;
    automata = automata;
    automata_type = automata_type;
    location = location;
    channels = channels;
    processes = processes;
    syntax_loc = syntax_loc;
    nmult = nmult;
    nchans = nchans;
    f = f;
    instvars = instvars
  }

and  translate_automata chans_env automata_type process_type automata =
  try
    let location =
      match automata.location with 
        None -> !current_location
      | Some location -> location
    in
    let chans_env =
      List.fold_right
        (fun (name, channel) env ->
          Env.add_value channel.chan_in_ident
            { val_type = channel.chan_type;
            val_kind = Val_channel location } env)
      automata.channels 
        (add_value automata.automata automata.automata_type chans_env)
    in
    let old_location = !current_location in
    current_location := location;
    let (decls,processes) =
      List.fold_left (fun (decls,processes) (join,process) ->
          let (process_exp,thread) = 
            make_process_def chans_env 
              (automata.automata,automata.f,automata.instvars)
            (join, process)
          in
          let thread =
            match thread with
              None ->
                if simple_thread chans_env process then
                  Some (-1)
                else
                  None
            | Some (name,reply_name) -> 
                Some
                  (try List.assoc name automata.channels
                  with
                    Not_found -> failwith "fill_automata"
                ).chan_ord
          in
          (mkpat (Tpat_var join.join_id)
            process_type,process_exp)::decls,
          (join,process_exp,thread)::processes)
      ([],[]) automata.processes
    in
    current_location := old_location;
    let (procs_env : Env.t) =
      List.fold_right
        (fun (join, process,thread) env ->
          add_value join.join_id process_type env)
      processes chans_env
    in
    let structure = 
      Janalyse.structure procs_env 
        automata.nmult automata.nchans automata.f processes in
    let trigger_list = List.map (fun (join,process,thread) -> 
          let path = Path.Pident join.join_id in
          let desc = Env.find_value path procs_env in
          mkexp  (Texp_ident (path, desc))
          process.exp_type procs_env
      ) processes
    in                     
(* one function to build the automata *)
    (fun exp ->
        mksequence
          (mklet decls
            (join_apply "fill_automata"
              [(ident_to_tident automata.automata procs_env);
              mktuple trigger_list procs_env;
              structure
            ]
              chans_env)
          chans_env)
        exp
          chans_env),
(* one function to build the init process *)
    (fun exp ->
        match automata.obj_init with
          None -> exp
        | Some p ->
            mksequence 
              (join_apply "agent_thread"
                (match automata.location with
                  None ->
                    [mkint 0 chans_env;
                      let (process, _, _, _) = 
                        type_process chans_env p ([],[],true) in
                      mkprocedure process chans_env
                    ]
                | Some location ->
                    [ident_to_tident location chans_env;
                      let old_location = !current_location in
                      current_location := location;
                      let (process, _, _, _) = 
                        type_process chans_env p ([],[],true) in
                      current_location := old_location;
                      mkprocedure process chans_env
                    ]) chans_env)
            exp
            chans_env
    ) 
  with
  | Failure s ->
      Printf.printf "f1<%s>" s; print_newline ();
      failwith s
      
      
and
  type_defs env syntax_loc automata_list =
  try
    begin_def ();  
    let automata_type = mkmodty env "Join" "automata" in
    let process_type = mkmodty env "Join" "process" in
    let class_type = mkmodty env "Join" "classe" in
    let automata_list = List.map (compile_automata env) automata_list
    in
(* the type of the resulting tuple of the whole definition *)
(*    let tuple_type =
      newty
        (Ttuple (List.flatten
            (List.map (fun a ->
                (List.map (fun (name, channel) -> channel.chan_type) 
                  a.channels))
            automata_list)))
in
  *)
    let chans_env =
      List.fold_right
        (fun  a env ->
          let location =
            match a.location with 
              None -> !current_location
            | Some location -> location
          in
          let env = add_value a.automata  automata_type env in
          match a.obj_name with
            None ->
              List.fold_right
                (fun (name, channel) env ->
                  Env.add_value channel.chan_in_ident
                    { val_type = channel.chan_type;
                    val_kind = Val_channel location }  env)
              a.channels env
          | Some obj ->
              add_value obj.obj_in_ident a.automata_type 
                (add_value obj.obj_class class_type env)
      ) automata_list env
    in
    
    let result_exp  = mktuple
        (List.flatten
          (List.map (fun a ->
              match a.obj_name with
                None ->
                  (List.map (fun (name, channel) -> 
                        ident_to_tident channel.chan_in_ident
                          chans_env
                    ) 
                    a.channels)
              | Some obj ->
                  [ident_to_tident obj.obj_in_ident chans_env]
          )
          automata_list)) chans_env
    in 
    
    let result_pat = 
      mkpattuple (List.flatten
          (List.map (fun a -> 
              match a.obj_name with
                None ->
                  List.map
                    (fun (name, channel) ->
                      mkpat (Tpat_var channel.chan_ext_ident) 
                      channel.chan_type)
                  a.channels
              | Some obj ->
                  [mkpat (Tpat_var obj.obj_ext_ident) a.automata_type]
          )
          automata_list))    
    
    in
    
    let auto_comps = List.map (
        translate_automata chans_env automata_type process_type) 
      automata_list
    in
          
    let exp =
      List.fold_left 
        (fun exp (_, init_exp) -> init_exp exp)
      result_exp auto_comps
    in
      
    let exp =
      List.fold_left 
        (fun exp (create_exp, _) -> create_exp exp)
      exp auto_comps
    in
    
    end_def();

(* generalization *)
    List.iter (fun a ->
        List.iter
          (fun (join, process) ->
            let tys = ref [] in
            List.iter
              (fun ty ->
                let newtys = ref [] in
                let rec f ty =
                  if List.memq ty !tys then make_nongen ty
                  else 
                  if not (List.memq ty !newtys) then
                    begin
                      newtys := ty :: !newtys; 
                      iter_type_expr f ty; () end
                in
                List.iter f ty; tys := !newtys @ !tys)
            join.join_types)
        a.processes; 
        List.iter (fun (name, channel) -> 
            generalize channel.chan_type) a.channels;    
    ) automata_list;
    let idents = List.map (fun a ->
          (a.location,
            match a.obj_name with 
              None ->
                List.map
                  (fun (name, channel) ->
                    channel.chan_ext_ident, channel.chan_type
                ) a.channels
            | Some obj -> 
                [obj.obj_ext_ident, a.automata_type]
          )) automata_list
    in
    check_multiple_define syntax_loc idents;
    
    let create_automata a =
      let chan_list =
        (mkpat (Tpat_var a.automata) automata_type):: 
        (List.map
            (fun (name, channel) ->
              mkpat (Tpat_var channel.chan_in_ident) 
              channel.chan_type)
          a.channels)
      in
      let indexes = mktuple (List.map (fun i-> mkint i env)
          (Array.to_list a.f)) env in
      let syncs = mktuple (List.map (fun (name,channel) -> 
              channel.chan_index <- a.f.(channel.chan_ord);
              match channel.chan_sync with
                None -> mkint 0 env
              | Some _ -> mkint 1 env
          ) a.channels) env in
      let args = [ 
          mkint a.nchans env;
          indexes;
          syncs
        ] in
      let loc_arg = match a.location with
          None -> mkint 0 env
        | Some location -> ident_to_tident location env 
      in
      let obj_arg = match a.obj_name with
          None -> mkint 0 env
        | Some obj -> ident_to_tident obj.obj_class chans_env
      in
      (mkpattuple chan_list,
        join_apply "create_empty_automata" (loc_arg::obj_arg::args) env
      )
    in      
    
    let exp = mklet (List.map create_automata automata_list) exp env in
    let tdef_env = List.fold_left 
        (fun env (_,loc_idents) ->
          List.fold_left
            (fun env (ident,idtype) -> add_value ident idtype env) env loc_idents)
      env idents
    in
    {
      tdef_defs = [ result_pat, exp] ;
      tdef_idents = idents;
      tdef_env = tdef_env;
    }
  with
    Not_found ->
      failwith "in type_defs"
  | Failure s ->
      Printf.printf "type_defs<%s>" s; print_newline ();
      failwith s
      
      (****
  
  let (c1,c2, ..., cn) = ???,
    let (automata1, c11,c21, ...,cm1) =
      create_empty_automata[_in location] nprocs nchans indexes syncs
    in
  
  
      ****)

and make_process_def env (automata,f,instvars) (join, process) =
  let new_env =
    List.fold_right
      (fun (id, (ty,vl)) env ->
        Env.add_value id {val_type = ty; val_kind = vl} env)
    join.join_vars env
  in
  begin_def (); 
  let (process_exp, replies, traps, thread) =
    type_process_and_trap_exc new_env process (join.join_replies,[],true)
  in
  end_def ();
  List.iter
    (fun (name, id) ->
      if not (List.mem name replies) then
        error process.pexp_loc ("No reply to channel " ^ name))
  join.join_replies;
  mkexp
    (Texp_function
      (let cases = [(match join.join_args with
              [ord, pat] -> pat
            | list ->
                let sorted_args =
                  Sort.list (fun x y -> 
                      (f.((fst x).chan_ord) land 63) <= 
                        (f.((fst y).chan_ord) land 63))
                  join.join_args
                in
                mkpattuple (List.map (fun (ord, pat) -> pat) sorted_args)),
          process_exp]
      in
      Parmatch.check_partial (!synthetic_loc) cases;
      cases
    ))
  (mkmodty env "Join" "process") env,
  thread

and type_process_and_trap_exc env proc context = 
  let (exp,replies,traps,thread) = type_process env proc context 
  in
  match traps with
    [] -> (* pas de reponses: inutile de faire un trap *)
      (exp,replies,traps,thread)
  | traps ->
(* des reponses : on place un handler d'exceptions autour du
          process dans lequel on expedie l'exception au canaux synchrones
          qui aurait du recevoir une reponse, puis on kill le thread *)
      let exn_id = Ident.create "exception" in
      let exn_type = instance Predef.type_exn in
      let new_env = add_value exn_id exn_type env in 
      (mkexp (Texp_try(exp,[
              mkpat (Tpat_var exn_id) exn_type,
              join_apply "send_exception" 
                [ident_to_tident exn_id new_env;
                mktuple (List.map (fun reply_name ->
                      ident_to_tident reply_name new_env
                  ) traps ) new_env 
              ] new_env
            ])) exp.exp_type env,
        replies, traps, thread)

(*
  Pour diminuer le nombre de threads lances, on peut ne pas forker 
  les threads non calculatoires, en posant thread = Some (-1).
  Toutefois, il faut noter que souvent, ces threads se terminent par
  un Join.exit ().
  L'autre solution est de ne forker que si necessaire, ie quand le scheduler
  ordonne la preemption du thread, ou que l'on fait un get_thread_stub.
  Et pourquoi pas les deux en meme temps ?

  En gros, on peut utiliser le <sync> pour reprendre le thread courant
  et l'utiliser a nouveau. Pour cela, il faut que trigger_def prenne la
  main immediatement, ce qui ne peut se faire que s'il est appele avec 
  un wrapper autour.

  Quelles criteres pour ces treads:
    Pas de reply
  
proc = { i:=!i+1; } | { k(j) }
 ---> i:=!i+1;k j;exit

Que faire des sends asynchrones terminaux ? Ils peuvent faire exit d'abord... 
mais au risque de vraiment tuer le thread !
*)

and simple_thread env proc =
(* toutes les operations qu'il contient sont simples *)
  let rec simple_thread proc =
    match proc.pexp_desc with
      Pexp_null -> true
    | Pexp_par list -> List.for_all simple_thread list
    | Pexp_call (e1,e2,_) -> not_computed env e1 && not_computed env e2
    | Pexp_sequence (e1,e2) -> not_computed env e1 && not_computed env e2
    | Pexp_let (_,list,p)     -> simple_thread p &&
        List.for_all (function (pat,e) -> not_computed env e) list
    | Pexp_def (_,p) -> simple_thread p
    | Pexp_ifthenelse (e1,p,option)  -> option_simple_thread option &&
        not_computed env e1 && simple_thread p
    | Pexp_location _
    | Pexp_reply _
    |  _ -> false
  
  and option_simple_thread = function
      None -> true
    | Some e -> simple_thread e
  in
  simple_thread proc

and type_process env proc (channels, replies, autonomep) =
  try
    match proc.pexp_desc with
      Pexp_null ->
        (join_apply "exit" [mkunit env] env,
          replies,[],None)
    | Pexp_par list ->
        let (sends,replys,threads,thread) =
          match List.fold_left (fun (s,r,a) p ->
                match p with
                  { pexp_desc = Pexp_call (e1,e2,_) } 
                    when not_computed  env e1 &&  not_computed env e2
                  -> p::s , r, a
                | { pexp_desc = Pexp_reply (e,_) } when not_computed env e
                  -> s, p::r,a
                | _ -> s , r, p::a) ([],[],[]) list
          with
            [],[],[] -> error proc.pexp_loc "{}|{} not accepted"
          | p::s , [], [] ->  s, [], [], p
          | s, p::r, [] -> s, r, [], p
          | s, r, p::t -> s, r, t, p
        in
        let proc_sends =
          List.map 
            (fun p ->
              let (p,_,_,_) = type_process env p (channels, replies, false)
              in p
          ) sends 
        in
        let proc_replies =
          List.fold_left 
            (fun (ps,replies,all_traps) p -> 
              let (p,replies,traps,_) = 
                type_process env p (channels,replies,false)
              in
              p::ps, replies, traps@all_traps
          ) (proc_sends,replies,[]) replys
        in
        let (proc_threads,replies, all_traps) =
          List.fold_left 
            (fun (ps,replies,all_traps) p ->
              let (p,replies,traps,thread) = 
                type_process_and_trap_exc env p (channels,replies,true)
              in
              let p =
                match thread with
                  None ->
                    join_apply "launch" [mkprocedure p env] env
                | Some (name,reply_name) ->
                    join_apply "reuse_thread"
                      [ident_to_tident reply_name env;
                      mkprocedure p env]
                      env
              in
              p::ps, replies, traps@all_traps
          ) proc_replies threads
        in
(* le thread principal *)
        let (p,replies,traps,thread) = 
          (match all_traps with
              [] -> type_process (* pas de replies supprimees *)
            | _ ->
                type_process_and_trap_exc)
          env thread (channels,replies,true)
        in
        (List.fold_left (fun exp_end p -> mksequence p exp_end env) 
          p proc_threads), 
        replies, traps@all_traps ,thread
    
    | Pexp_let (rec_flag, spat_sexp_list, sbody) ->
        let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
        let (body, replies, traps, thread) = type_process new_env sbody 
            (channels,replies,true) in
        ({exp_desc = Texp_let (rec_flag, pat_exp_list, body);
            exp_loc = proc.pexp_loc; 
            exp_type = body.exp_type; 
            exp_env = env},
          replies, traps, thread)
    | Pexp_sequence (sexp1,sexp2) ->
        let exp1 = type_statement env sexp1 in
        let (exp2, replies, traps, thread) = type_process env sexp2 
            (channels,replies,true) in
        ({exp_desc = Texp_sequence (exp1, exp2);
            exp_loc = proc.pexp_loc; 
            exp_type = exp2.exp_type; 
            exp_env = env},
          replies, traps, thread)
    
    | Pexp_location (loc_list, sbody) ->
        let tloc = 
          (*  (loc_defs,chan_defs,threads_exp,loc_env,new_env) = *)
          type_locations  env proc.pexp_loc loc_list
        in
        let (body, replies, traps, thread) =  
          type_process tloc.tloc_fullenv sbody (channels,replies,true)
        in
        mklet tloc.tloc_locs
          (mklet tloc.tloc_defs
            (mksequence tloc.tloc_inits body tloc.tloc_fullenv)
          tloc.tloc_locenv)
        env,
        replies, traps, thread
    | Pexp_def (spat_sexp_list, sbody) ->
        let (pat_exp_list, new_env) =
          type_def env proc.pexp_loc spat_sexp_list
        in
        let (body, replies, traps, thread) = type_process new_env sbody
            (channels,replies,true) in
        ({exp_desc = Texp_let (Nonrecursive, pat_exp_list, body);
            exp_loc = proc.pexp_loc;
            exp_type = body.exp_type; 
            exp_env = env},
          replies, traps, thread)
    | Pexp_ifthenelse (scond, p, q) ->
        let cond = type_expect env scond (instance Predef.type_bool) in
        let (proc_p, repl_p, traps_p, thread_p) = 
          type_process env p (channels,replies,true) in
        let (proc_q, repl_q, traps_q, thread_q) = 
          match q with
            Some q -> type_process env q (channels,replies,true) 
          | None ->
              (join_apply "exit" [mkunit env] env,
                replies, [], None)
        in
        if List.length repl_p <> List.length repl_q ||
          List.exists (fun e -> not (List.mem e repl_p)) repl_q then
          error proc.pexp_loc "Not the same replies in IF alternatives";
        ({exp_desc = Texp_ifthenelse (cond, proc_p, Some proc_q);
            exp_loc = proc.pexp_loc; exp_type = proc_p.exp_type;
            exp_env = env},
          repl_p, traps_p,
          if thread_p = thread_q then thread_p else None)
    | Pexp_call (e1, e2,_) ->
        let exp =
          let channel = type_exp env e1 in
          let (tyl) =
            try
              filter_async_channel env (instance channel.exp_type)
            with
              Failure s -> 
                raise (Error (proc.pexp_loc, 
                    SimpleTypeError (s,channel.exp_type)))
          in
          let arg_exp = type_expect env e2 tyl in
          mkexp (Texp_apply (channel,[arg_exp])) (instance Predef.type_unit) env 
        in
        (if autonomep then
            mkexp(Texp_sequence(exp,
                join_apply "exit" [mkunit env] env
              )) (instance Predef.type_unit) env
          else
            exp),
        replies, [], None
    
    | Pexp_match  (sarg, (spat,sexp) :: otherlist) ->
        let arg = type_exp env sarg in
        let (pat, ext_env) = type_pattern env spat in
        unify_pat env pat arg.exp_type;
        let (exp, repl_p, traps_p, thread_p) = 
          type_process ext_env sexp (channels,replies,true) 
        in
        let thread_p = ref thread_p in
        let cases = (pat, exp)::(List.map
              (fun (spat, sexp) ->
                let (pat, ext_env) = type_pattern env spat in
                unify_pat env pat arg.exp_type;
                let (exp, repl_q, traps_q, thread_q) = 
                  type_process ext_env sexp (channels,replies,true) 
                in
                if List.length repl_p <> List.length repl_q ||
                  List.exists (fun e -> not (List.mem e repl_p)) repl_q then
                  error proc.pexp_loc "Not the same replies in MATCH alternatives";
                if !thread_p <> thread_q then thread_p := None;
                (pat, exp)
            ) otherlist)
        in
        Parmatch.check_unused cases;
        if Parmatch.check_partial sexp.pexp_loc cases then
(* le cas partiel n'est pas autorise pour l'instant ...  *)
          error proc.pexp_loc "Partial matching not authorised in processes";
        ({exp_desc = Texp_match (arg, cases); exp_loc = sexp.pexp_loc;
            exp_type = exp.exp_type; exp_env = env}, repl_p, traps_p, !thread_p)

(* reply( let defs in e ) -> let defs in reply e *)
    | Pexp_reply ( 
        { pexp_desc = Pexp_let (rec_flag, spat_sexp_list, sbody);
          pexp_loc = exp_loc
        }, name) ->
        type_process env 
          { pexp_desc = Pexp_let(rec_flag, spat_sexp_list, 
            { pexp_desc = Pexp_reply(sbody,name); 
              pexp_loc = proc.pexp_loc});
          pexp_loc = exp_loc
        } (channels,replies,autonomep)

(* reply( let defs in e ) -> let defs in reply e *)
    | Pexp_reply ( 
        { pexp_desc = Pexp_def (spat_sexp_list, sbody);
          pexp_loc = exp_loc
        }, name) ->
        type_process env 
          { pexp_desc = Pexp_def(spat_sexp_list, 
            { pexp_desc = Pexp_reply(sbody,name); 
              pexp_loc = proc.pexp_loc});
          pexp_loc = exp_loc
        } (channels,replies,autonomep)

(* reply( if c then e1 else e2 ) -> if c then reply e1 else reply e2 *)
    | Pexp_reply ( 
        { pexp_desc = Pexp_ifthenelse(cond,e1,option);
          pexp_loc = exp_loc
        }, name) ->
        type_process env 
          { pexp_desc = 
          Pexp_ifthenelse(cond,  
            { pexp_desc = Pexp_reply(e1,name); 
              pexp_loc = proc.pexp_loc},
            Some { pexp_desc = Pexp_reply(
                (match option with
                    Some e2 -> e2
                  | None -> { pexp_desc = 
                        Pexp_construct (Lident "()", None, false);
                        pexp_loc = exp_loc}
                ),name);
              pexp_loc = proc.pexp_loc}
          );
          pexp_loc = exp_loc
        } (channels,replies,autonomep)

(* reply( if c then e1 else e2 ) -> if c then reply e1 else reply e2 *)
    | Pexp_reply ( 
        { pexp_desc = Pexp_match(cond,list);
          pexp_loc = exp_loc
        }, name) ->
        type_process env 
          { pexp_desc = 
          Pexp_match(cond, List.map (fun (pat,e) ->
                pat, { pexp_desc = Pexp_reply(e,name); 
                  pexp_loc = proc.pexp_loc}) list);
          pexp_loc = exp_loc
        } (channels,replies,autonomep)

(* reply( s << n >> ) -> tailcall *)
    | Pexp_reply ( sexp,name) ->
        type_reply sexp name proc (channels, replies, autonomep) env
    
    | _ -> assert false
  
  with
    Not_found ->
      failwith "not found in process"

and type_reply sexp name proc (channels, replies, autonomep) env =
  let name,reply_name,lid = 
    match (name,channels) with
      (None,[]) -> 
        error proc.pexp_loc "Reply with no synchronous channel"
    | (None,[name,reply_name]) ->
        (name,reply_name,(Longident.Lident name))
    | (None,_) -> error proc.pexp_loc 
          "Anonymous reply with several synchronous channels"
    | (Some name,_) ->
        (name,(
            try List.assoc name channels
            with
              Not_found -> 
                error proc.pexp_loc
                  ("Reply to channel external from def "^name)
          ),(Longident.Lident name) )
  in
  if List.mem name replies then
    error proc.pexp_loc
      ("Two replies to the same channel "^name);
  
  let (path, desc) = 
    try
      Env.lookup_value lid env 
    with
      Not_found ->
        raise (Error (sexp.pexp_loc, Unbound_value lid))
  in
  begin
    match path with
      Path.Pident ident -> 
        type_reply2 sexp name proc (channels, replies, autonomep) env ident desc reply_name
    | _ -> error proc.pexp_loc "Bad name in Reply"
  end

and  type_reply2 sexp name proc (channels, replies, autonomep) env ident desc reply_name =
  let (ty1, ty) = try filter_arrow env desc.val_type with
    | _ -> raise (Error(proc.pexp_loc, Expr_type_clash
              (let t2 = newty (Tarrow(newvar(), newvar())) in
              [(desc.val_type,desc.val_type); (t2,t2)])))
  in
  match sexp with
    { pexp_desc = Pexp_apply({ pexp_desc = Pexp_ident lid} as e1,[e2]); 
      pexp_loc = exp_loc } ->
(* tail call *)
      let (path,desc) = try Env.lookup_value lid env with
          Not_found -> raise (Error (sexp.pexp_loc, Unbound_value lid))
      in
      begin
        match desc.val_kind with
          Val_channel _  -> begin
              Debug.print0 'f' "Tailcall\n";
              let id = ident_to_tident reply_name env in
              let channel = type_exp env e1 in
              let (tyl, rty) = try
                  filter_sync_channel env (instance channel.exp_type)
                with Failure s -> 
                    raise (Error (proc.pexp_loc, 
                        SimpleTypeError (s,channel.exp_type)))
              in
              let arg_exp = type_expect env e2 tyl in
              let exp =
                {exp_desc = Texp_apply
                    (type_exp env (mkmodpident "Join" 
                        (match desc.val_kind with
                          Val_channel location when 
                          Ident.same location !current_location ->
                            "unsafe_tail_synchrone"
                        | _ -> "safe_tail_synchrone")),
                    channel::arg_exp::[id]);
                  exp_loc = exp_loc; 
                  exp_type = rty; 
                  exp_env = env}
              in
              (try unify env ty exp.exp_type with
                  Unify trace -> 
                    raise(Error(proc.pexp_loc, Expr_type_clash(trace))));
              { exp_desc = exp.exp_desc;
                exp_loc = exp.exp_loc;
                exp_type = ty; 
                exp_env = env}
                , name::replies, [reply_name],
              (Some (name,reply_name))
            end
        
        | _ -> (* apply function : normal case *)
            let id = ident_to_tident reply_name env in
            let exp = type_expect env sexp ty in
            let apply_Jrt = 
              join_apply 
                (if autonomep then "reply_to_and_exit"
                else "reply_to")
              [id; exp] env
            in
            apply_Jrt, name::replies, [reply_name],
            (Some (name,reply_name))
      end
  
  | _ -> (* value : normal case *)
      let id = ident_to_tident reply_name env in
      let exp = type_expect env sexp ty in
      let apply_Jrt = 
        join_apply 
          (if autonomep then "reply_to_and_exit"
          else "reply_to")
        [id; exp] env
      in
      apply_Jrt, name::replies, [reply_name],
      (Some (name,reply_name))

and
  type_locations env syntax_loc location_list =
  let location_type = mkmodty env "Join" "location" in
  let location_list =
    List.map 
      (fun (loc_name,modules,automatas,pexp) -> 
        let ident = Ident.create loc_name in
        (ident,modules,automatas,pexp)
    ) location_list
  in
  let location_env = (* env + locations *)
    List.fold_left 
      (fun env (ident,modules,automatas,exp) ->
        Env.add_value ident 
          {val_type = location_type; 
          val_kind = Val_reg} env
    ) env location_list
  in
(*  let (decls,idents,automata_env) =  *)
  let tdef = type_defs location_env syntax_loc 
      (List.flatten
        (List.map
          (fun (ident,modules,automatas,exp) ->
            List.map (fun automata ->
                (Some ident, automata,syntax_loc)
            ) automatas)
        location_list))
  in
  let tloc_locs = List.map
      (fun (ident,modules,automatas,exp) ->
        (mkpat (Tpat_var ident) location_type),
        (join_apply "create_agent"
            [mkexp Texp_thismodule (join_type "module_data" env) env;
            type_exp env modules
          ]
            env)
    ) location_list in
  let tloc_defs =   tdef.tdef_defs in
  let tloc_inits = 
    let start_thread (location,modules,automatas,process) =
      let new_env =
        add_value location location_type
          (List.fold_left
            (fun env (l,chans) ->
              match l with
                Some l when Ident.same l location ->
                  List.fold_left 
                    (fun env (ident,idtype) ->
                      add_value ident  idtype env) env chans
              | _ -> env) tdef.tdef_env tdef.tdef_idents)
      in
      join_apply "agent_thread"
        [ident_to_tident location new_env;
        let old_location = !current_location in
        current_location := location;
        let (process, _, _, _) = type_process new_env process ([],[],true) in
        current_location := old_location;
        mkprocedure process new_env
      ] new_env
    in
    List.fold_left
      (fun exp ((location,modules,automatas,process) as tuple) ->
        match process.pexp_desc with
          Pexp_null -> exp
        | _ -> mksequence  exp (start_thread tuple) tdef.tdef_env)
    (start_thread (List.hd location_list))
    (List.tl location_list)
  in
  let tloc_locenv = location_env in
  let tloc_fullenv = tdef.tdef_env in
  { 
    tloc_defs = tloc_defs;
    tloc_locs = tloc_locs;
    tloc_inits = tloc_inits;
    tloc_locenv = tloc_locenv;
    tloc_fullenv = tloc_fullenv;
  }
    
(* Typing of toplevel def bindings *)

let type_loc_binding env location loc_list =
  Typetexp.reset_type_variables (); 
  type_locations env location loc_list
    ;;
let type_join_binding env location ddefs =
  Typetexp.reset_type_variables (); 
  type_def env location ddefs
    ;;

let _ = 
  Typecore.type_defs := type_defs;
  Typecore.type_locations := type_locations;
  Typecore.type_process := type_process
  
  